# This file is part of MyPaint.
# Copyright (C) 2013-2018 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""Dockable panel showing options for the current mode"""

## Imports

from __future__ import division, print_function
import logging

from .toolstack import SizedVBoxToolWidget, TOOL_WIDGET_NATURAL_HEIGHT_SHORT
import lib.xml
from lib.gettext import C_

from lib.gibindings import Gtk

logger = logging.getLogger(__name__)


## Class defs

class ModeOptionsTool (SizedVBoxToolWidget):
    """Dockable panel showing options for the current mode

    This panel has a title and an icon reflecting the current mode, and
    displays its options widget if it has one: define an object method named
    ``get_options_widget()`` returning an arbitrary GTK widget. Singletons work
    well here, and are encouraged. ``get_options_widget()`` can also return
    `None` if the mode is a temporary mode which has no sensible options. In
    this case, any widget already displayed will not be replaced, which is
    particularly appropriate for modes which only persist for the length of
    time the mouse button is held, and stack on top of other modes.
    """

    ## Class constants

    SIZED_VBOX_NATURAL_HEIGHT = TOOL_WIDGET_NATURAL_HEIGHT_SHORT

    tool_widget_icon_name = "mypaint-options-symbolic"
    tool_widget_title = C_(
        "options panel: tab tooltip: title",
        "Tool Options",
    )
    tool_widget_description = C_(
        "options panel: tab tooltip: description",
        "Specialized settings for the current editing tool",
    )

    __gtype_name__ = 'MyPaintModeOptionsTool'

    OPTIONS_MARKUP = C_(
        "options panel: header",
        "<b>{mode_name}</b>",
    )
    NO_OPTIONS_MARKUP = C_(
        "options panel: body",
        "<i>No options available</i>",
    )

    ## Method defs

    def __init__(self):
        """Construct, and connect internal signals & callbacks"""
        SizedVBoxToolWidget.__init__(self)
        from gui.application import get_app
        self._app = get_app()
        self._app.doc.modes.changed += self._modestack_changed_cb
        self.set_border_width(3)
        self.set_spacing(6)
        # Placeholder in case a mode has no options
        label = Gtk.Label()
        label.set_markup(self.NO_OPTIONS_MARKUP)
        self._no_options_label = label
        # Container for an options widget exposed by the current mode
        self._mode_icon = Gtk.Image()
        label = Gtk.Label()
        label.set_text("<options-label>")
        self._options_label = label
        label.set_alignment(0.0, 0.5)
        label_hbox = Gtk.HBox()
        label_hbox.set_spacing(3)
        label_hbox.set_border_width(3)
        label_hbox.pack_start(self._mode_icon, False, False, 0)
        label_hbox.pack_start(self._options_label, True, True, 0)
        align = Gtk.Alignment.new(0.5, 0.5, 1.0, 1.0)
        align.set_padding(0, 0, 0, 0)
        align.set_border_width(3)
        self._options_bin = align
        self.pack_start(label_hbox, False, False, 0)
        self.pack_start(align, True, True, 0)
        self.connect("show", lambda *a: self._update_ui())
        # Fallback
        self._update_ui_with_options_widget(
            self._no_options_label,
            self.tool_widget_title,
            self.tool_widget_icon_name,
        )

    def _modestack_changed_cb(self, modestack, old, new):
        """Update the UI when the mode changes"""
        self._update_ui()

    def _update_ui(self):
        """Update the UI to show the options widget of the current mode"""
        mode = self._app.doc.modes.top
        self._update_ui_for_mode(mode)

    def _update_ui_for_mode(self, mode):
        # Get the new options widget
        try:
            get_options_widget = mode.get_options_widget
        except AttributeError:
            get_options_widget = None
        if get_options_widget:
            new_options = get_options_widget()
        else:
            new_options = self._no_options_label
        if not new_options:
            # Leave existing widget as-is, even if it's the default.
            # XXX maybe we should be doing something stack-based here?
            return
        icon_name = mode.get_icon_name()
        name = mode.get_name()
        self._update_ui_with_options_widget(new_options, name, icon_name)

    def _update_ui_with_options_widget(self, new_options, name, icon_name):
        old_options = self._options_bin.get_child()
        logger.debug("name: %r, icon name: %r", name, icon_name)
        if name:
            markup = self.OPTIONS_MARKUP.format(
                mode_name=lib.xml.escape(name),
            )
            self._options_label.set_markup(markup)
        if icon_name:
            self._mode_icon.set_from_icon_name(
                icon_name,
                Gtk.IconSize.SMALL_TOOLBAR,
            )
        # Options widget: only update if there's a change
        if new_options is not old_options:
            if old_options:
                old_options.hide()
                self._options_bin.remove(old_options)
            self._options_bin.add(new_options)
            new_options.show()
        self._options_bin.show_all()



# -*- coding: utf-8 -*-
# This file is part of MyPaint.
# Copyright (C) 2012-2018 by the MyPaint Development Team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


"""Hue/Relative chroma/Luma adjuster widgets, with an editable gamut mask.
"""

from __future__ import division, print_function

import math
from copy import deepcopy
import re
import os.path

from lib.gibindings import Gtk
from lib.gibindings import Gdk
import cairo

from .adjbases import ColorManager
from .adjbases import ColorAdjuster
from .adjbases import HueSaturationWheelMixin
from .adjbases import HueSaturationWheelAdjuster
from .sliders import HCYLumaSlider
from .combined import CombinedAdjusterPage
from lib.color import RGBColor
from lib.color import HCYColor
from lib.color import HSVColor
import gui.uicolor
from .util import clamp
from lib.palette import Palette
import lib.alg as geom
from .paletteview import palette_load_via_dialog
from .paletteview import palette_save_via_dialog
from lib.gettext import C_

from lib.pycompat import xrange


PREFS_MASK_KEY = "colors.hcywheel.mask.gamuts"
PREFS_ACTIVE_KEY = "colors.hcywheel.mask.active"
MASK_EDITOR_HELP_URI = C_(
    "Online help pages",
    u"https://github.com/mypaint/mypaint/wiki/"
    u"v1.2-HCY-Wheel-and-Gamut-Mask-Editor"
)


class MaskableWheelMixin(object):
    """Provides wheel widgets with maskable areas.

    For use with implementations of `HueSaturationWheelAdjusterMixin`.
    Concrete implementations can be masked so that they ignore clicks outside
    certain color areas. If the mask is active, clicks inside the mask
    shapes are treated as normal, but clicks outside them are remapped to a
    point on the nearest edge of the nearest shape. This can be useful for
    artists who wish to plan the color gamut of their artwork in advance.

    http://gurneyjourney.blogspot.com/2011/09/part-1-gamut-masking-method.html
    http://gurneyjourney.blogspot.com/2008/01/color-wheel-masking-part-1.html

    """

    # Class-level variables: drawing constants etc.
    min_shape_size = 0.15  # Smallest useful shape: fraction of radius

    # Instance variables (defaults / documentation)
    __mask = None
    mask_toggle = None  # Gtk.ToggleAction controlling whether the mask is used
    mask_observers = None  # List of no-argument mask change obs'r callbacks

    def __init__(self):
        """Instantiate instance vars and bind actions.
        """
        self.__mask = []
        self.mask_observers = []
        action_name = "wheel%s_masked" % (id(self),)
        self.mask_toggle = Gtk.ToggleAction(
            action_name,
            C_(
                "Color Wheels: activity toggle: action title",
                u"Gamut Mask Active",
            ),
            C_(
                "Color Wheels: activity toggle: action tooltip",
                u"Limit your palette for specific moods using a gamut mask.",
            ),
            None
        )
        self.mask_toggle.connect("toggled", self.__mask_toggled_cb)

    def __mask_toggled_cb(self, action):
        active = action.get_active()
        prefs = self.get_prefs()
        prefs[PREFS_ACTIVE_KEY] = active
        self.queue_draw()

    def set_color_manager(self, manager):
        """Sets the color manager, and reads an initial mask from prefs.

        Extends `ColorAdjuster`'s implementation.

        """
        ColorAdjuster.set_color_manager(self, manager)
        prefs = self.get_prefs()
        mask_flat = prefs.get(PREFS_MASK_KEY, None)
        mask_active = prefs.get(PREFS_ACTIVE_KEY, False)
        if mask_flat is not None:
            self.set_mask(self._unflatten_mask(mask_flat))
            self.mask_toggle.set_active(mask_active)

    @staticmethod
    def _flatten_mask(mask):
        flat_mask = []
        for shape_colors in mask:
            shape_flat = [c.to_hex_str() for c in shape_colors]
            flat_mask.append(shape_flat)
        return flat_mask

    @staticmethod
    def _unflatten_mask(flat_mask):
        mask = []
        for shape_flat in flat_mask:
            shape_colors = [RGBColor.new_from_hex_str(s) for s in shape_flat]
            mask.append(shape_colors)
        return mask

    def set_mask_from_palette(self, pal):
        """Sets the mask from a palette.

        Any `palette.Palette` can be loaded into the wheel widget, and color
        names are used for distinguishing mask shapes. If a color name
        matches the pattern "``mask #<decimal-int>``", it will be associated
        with the shape having the ID ``<decimal-int>``.

        """
        if pal is None:
            return
        mask_id_re = re.compile(r'\bmask\s*#?\s*(\d+)\b')
        mask_shapes = {}
        for i in xrange(len(pal)):
            color = pal.get_color(i)
            if color is None:
                continue
            shape_id = 0
            color_name = pal.get_color_name(i)
            if color_name is not None:
                mask_id_match = mask_id_re.search(color_name)
                if mask_id_match:
                    shape_id = int(mask_id_match.group(1))
            if shape_id not in mask_shapes:
                mask_shapes[shape_id] = []
            mask_shapes[shape_id].append(color)
        mask_list = []
        shape_ids = sorted(mask_shapes.keys())
        for shape_id in shape_ids:
            mask_list.append(mask_shapes[shape_id])
        self.set_mask(mask_list)

    def set_mask(self, mask):
        """Sets the mask (a list of lists of `UIColor`s).
        """
        prefs = self.get_prefs()
        if mask is None:
            self.__mask = None
            self.mask_toggle.set_active(False)
            prefs[PREFS_MASK_KEY] = None
        else:
            self.mask_toggle.set_active(True)
            self.__mask = mask
            prefs[PREFS_MASK_KEY] = self._flatten_mask(mask)
        for func in self.mask_observers:
            func()
        self.queue_draw()

    def get_mask(self):
        """Returns the current mask.
        """
        return self.__mask

    def get_mask_voids(self):
        """Returns the current mask as a list of lists of (x, y) pairs.
        """
        voids = []
        if not self.__mask:
            return voids
        for shape in self.__mask:
            if len(shape) >= 3:
                void = self.colors_to_mask_void(shape)
                voids.append(void)
        return voids

    def colors_to_mask_void(self, colors):
        """Converts a set of colors to a mask void (convex hull).

        Mask voids are the convex hulls of the (x, y) positions for the
        colors making up the mask, so mask shapes with fewer than 3 colors
        are returned as the empty list.

        """
        points = []
        if len(colors) < 3:
            return points
        for col in colors:
            points.append(self.get_pos_for_color(col))
        return geom.convex_hull(points)

    def get_color_at_position(self, x, y, ignore_mask=False):
        """Converts an `x`, `y` position to a color.

        Ordinarily, this implementation uses any active mask to limit the
        colors which can be clicked on. Set `ignore_mask` to disable this
        added behaviour.

        """
        sup = HueSaturationWheelMixin
        if ignore_mask or not self.mask_toggle.get_active():
            return sup.get_color_at_position(self, x, y)
        voids = self.get_mask_voids()
        if not voids:
            return sup.get_color_at_position(self, x, y)
        isects = []
        for vi, void in enumerate(voids):
            # If we're inside a void, use the unchanged value
            if geom.point_in_convex_poly((x, y), void):
                return sup.get_color_at_position(self, x, y)
            # If outside, find the nearest point on the nearest void's edge
            for p1, p2 in geom.pairwise(void):
                isect = geom.nearest_point_in_segment(p1, p2, (x, y))
                if isect is not None:
                    d = math.sqrt((isect[0]-x)**2 + (isect[1]-y)**2)
                    isects.append((d, isect))
                # Above doesn't include segment ends, so add those
                d = math.sqrt((p1[0]-x)**2 + (p1[1]-y)**2)
                isects.append((d, p1))
        # Determine the closest point.
        if isects:
            isects.sort()
            x, y = isects[0][1]
        return sup.get_color_at_position(self, x, y)

    @staticmethod
    def _get_void_size(void):
        """Size metric for a mask void (list of x,y points; convex hull)
        """
        area = geom.poly_area(void)
        return math.sqrt(area)

    def _get_mask_fg(self):
        """Returns the mask edge drawing color as an rgb triple"""
        state = self.get_state_flags()
        style = self.get_style_context()
        c = style.get_color(state)
        return gui.uicolor.from_gdk_rgba(c).get_rgb()

    def _get_mask_bg(self):
        """Returns the mask area drawing color as an rgb triple."""
        state = self.get_state_flags()
        style = self.get_style_context()
        c = style.get_background_color(state)
        return gui.uicolor.from_gdk_rgba(c).get_rgb()

    def draw_mask(self, cr, wd, ht):
        """Draws the mask, if enabled and if it has any usable voids.

        For the sake of the editor subclass, this doesn't draw any voids
        which are smaller than `self.min_shape_size` times the wheel radius.

        """

        if not self.mask_toggle.get_active():
            return
        if self.__mask is None or self.__mask == []:
            return

        cr.save()

        radius = self.get_radius(wd=wd, ht=ht)
        cx, cy = self.get_center(wd=wd, ht=ht)
        cr.arc(cx, cy, radius+self.BORDER_WIDTH, 0, 2*math.pi)
        cr.clip()

        bg_rgb = self._get_mask_bg()
        fg_rgb = self._get_mask_fg()

        cr.push_group()
        cr.set_operator(cairo.OPERATOR_OVER)
        cr.set_source_rgb(*bg_rgb)
        cr.rectangle(0, 0, wd, ht)
        cr.fill()
        voids = []
        min_size = radius * self.min_shape_size
        for void in self.get_mask_voids():
            if len(void) < 3:
                continue
            size = self._get_void_size(void)
            if size >= min_size:
                voids.append(void)
        cr.set_source_rgb(*fg_rgb)
        for void in voids:
            cr.new_sub_path()
            cr.move_to(*void[0])
            for x, y in void[1:]:
                cr.line_to(x, y)
            cr.close_path()
        cr.set_line_width(2.0)
        cr.stroke_preserve()
        cr.set_operator(cairo.OPERATOR_SOURCE)
        cr.set_source_rgba(1, 1, 1, 0)
        cr.fill()
        cr.set_operator(cairo.OPERATOR_OVER)
        cr.pop_group_to_source()

        cr.paint_with_alpha(0.666)
        cr.restore()

    def paint_foreground_cb(self, cr, wd, ht):
        """Paints the foreground items: mask, then marker.
        """
        self.draw_mask(cr, wd, ht)
        HueSaturationWheelMixin.paint_foreground_cb(self, cr, wd, ht)


class HCYHueChromaWheelMixin(object):
    """Mixin for wheel-style adjusters to display the H+C from the HCY model.

    For use with implementations of `HueSaturationWheelAdjusterMixin`; make
    sure this mixin comes before it in the MRO.

    """

    def get_normalized_polar_pos_for_color(self, col):
        col = HCYColor(color=col)
        return col.c, col.h

    def color_at_normalized_polar_pos(self, r, theta):
        col = HCYColor(color=self.get_managed_color())
        col.h = theta
        col.c = r
        return col


class HCYHueChromaWheel (MaskableWheelMixin,
                         HCYHueChromaWheelMixin,
                         HueSaturationWheelAdjuster):
    """Circular mapping of the H and C terms of the HCY model."""

    STATIC_TOOLTIP_TEXT = C_(
        "HCY Color Wheel: tooltip",
        u"HCY hue and chroma.",
    )

    def __init__(self):
        """Instantiate, binding events.
        """
        MaskableWheelMixin.__init__(self)
        HueSaturationWheelAdjuster.__init__(self)
        self.connect("scroll-event", self.__scroll_cb)
        self.add_events(Gdk.EventMask.SCROLL_MASK)

    def __scroll_cb(self, widget, event):
        # Scrolling controls luma.
        d = self.SCROLL_DELTA
        directions = (Gdk.ScrollDirection.DOWN, Gdk.ScrollDirection.LEFT)
        if event.direction in directions:
            d *= -1
        col = HCYColor(color=self.get_managed_color())
        y = clamp(col.y+d, 0.0, 1.0)
        if col.y != y:
            col.y = y
            self.set_managed_color(col)
        return True


class HCYMaskEditorWheel (HCYHueChromaWheel):
    """HCY wheel specialized for mask editing."""

    ## Instance vars
    __last_cursor = None   # previously set cursor (determines some actions)
    # Objects which are active or being manipulated
    __tmp_new_ctrlpoint = None   # new control-point color
    __active_ctrlpoint = None   # active point in active_void
    __active_shape = None  # list of colors or None
    # Drag state
    __drag_func = None
    __drag_start_pos = None

    ## Class-level constants and variables
    # Specialized cursors for different actions
    __add_cursor = None
    __move_cursor = None
    __move_point_cursor = None
    __rotate_cursor = None
    # Constrain the range of allowable lumas
    __MAX_LUMA = 0.75
    __MIN_LUMA = 0.25

    # Drawing constraints and activity proximities
    __ctrlpoint_radius = 2.5
    __ctrlpoint_grab_radius = 10
    __max_num_shapes = 6   # how many shapes are allowed

    # Tooltip text. Is here a better way of explaining this? It obscures the
    # editor quite a lot.
    STATIC_TOOLTIP_TEXT = C_(
        "HCY Mask Editor Wheel: tooltip",
        u"Gamut mask editor. Click in the middle to create "
        u"or manipulate shapes, or rotate the mask using "
        u"the edges of the disc.",
    )

    def __init__(self):
        """Instantiate, and connect the editor events.
        """
        super(HCYMaskEditorWheel, self).__init__()

        self.connect("realize", self._realize_cb)
        self.connect("button-press-event", self.__button_press_cb)
        self.connect("button-release-event", self.__button_release_cb)
        self.connect("motion-notify-event", self.__motion_cb)
        self.connect("leave-notify-event", self.__leave_cb)
        self.add_events(
            Gdk.EventMask.POINTER_MOTION_MASK |
            Gdk.EventMask.LEAVE_NOTIFY_MASK)

    def _realize_cb(self, widget):
        display = self.get_window().get_display()

        self.__add_cursor = Gdk.Cursor.new_for_display(
            display, Gdk.CursorType.PLUS)
        self.__move_cursor = Gdk.Cursor.new_for_display(
            display, Gdk.CursorType.FLEUR)
        self.__move_point_cursor = Gdk.Cursor.new_for_display(
            display, Gdk.CursorType.CROSSHAIR)
        self.__rotate_cursor = Gdk.Cursor.new_for_display(
            display, Gdk.CursorType.EXCHANGE)

    def __leave_cb(self, widget, event):
        # Reset the active objects when the pointer leaves.
        if self.__drag_func is not None:
            return
        self.__active_shape = None
        self.__active_ctrlpoint = None
        self.__tmp_new_ctrlpoint = None
        self.queue_draw()
        self.__set_cursor(None)

    def __set_cursor(self, cursor):
        # Sets the window cursor, retaining a record.
        if cursor != self.__last_cursor:
            self.get_window().set_cursor(cursor)
            self.__last_cursor = cursor

    def __update_active_objects(self, x, y):
        # Decides what a click or a drag at (x, y) would do, and updates the
        # mouse cursor and draw state to match.

        assert self.__drag_func is None
        self.__active_shape = None
        self.__active_ctrlpoint = None
        self.__tmp_new_ctrlpoint = None
        self.queue_draw()  # yes, always

        # Possible mask void manipulations
        mask = self.get_mask()
        for mask_idx in xrange(len(mask)):
            colors = mask[mask_idx]
            if len(colors) < 3:
                continue

            # If the pointer is near an existing control point, clicking and
            # dragging will move it.
            void = []
            for col_idx in xrange(len(colors)):
                col = colors[col_idx]
                px, py = self.get_pos_for_color(col)
                dp = math.sqrt((x-px)**2 + (y-py)**2)
                if dp <= self.__ctrlpoint_grab_radius:
                    mask.remove(colors)
                    mask.insert(0, colors)
                    self.__active_shape = colors
                    self.__active_ctrlpoint = col_idx
                    self.__set_cursor(None)
                    return
                void.append((px, py))

            # If within a certain distance of an edge, dragging will create and
            # then move a new control point.
            void = geom.convex_hull(void)
            for p1, p2 in geom.pairwise(void):
                isect = geom.nearest_point_in_segment(p1, p2, (x, y))
                if isect is not None:
                    ix, iy = isect
                    di = math.sqrt((ix-x)**2 + (iy-y)**2)
                    if di <= self.__ctrlpoint_grab_radius:
                        newcol = self.get_color_at_position(ix, iy)
                        self.__tmp_new_ctrlpoint = newcol
                        mask.remove(colors)
                        mask.insert(0, colors)
                        self.__active_shape = colors
                        self.__set_cursor(None)
                        return

            # If the mouse is within a mask void, then dragging would move that
            # shape around within the mask.
            if geom.point_in_convex_poly((x, y), void):
                mask.remove(colors)
                mask.insert(0, colors)
                self.__active_shape = colors
                self.__set_cursor(None)
                return

        # Away from shapes, clicks and drags manipulate the entire mask: adding
        # cutout voids to it, or rotating the whole mask around its central
        # axis.
        alloc = self.get_allocation()
        cx, cy = self.get_center(alloc=alloc)
        radius = self.get_radius(alloc=alloc)
        dx, dy = x-cx, y-cy
        r = math.sqrt(dx**2 + dy**2)
        if r < radius*(1.0-self.min_shape_size):
            if len(mask) < self.__max_num_shapes:
                d = self.__dist_to_nearest_shape(x, y)
                minsize = radius * self.min_shape_size
                if d is None or d > minsize:
                    # Clicking will result in a new void
                    self.__set_cursor(self.__add_cursor)
        else:
            # Click-drag to rotate the entire mask
            self.__set_cursor(self.__rotate_cursor)

    def __drag_active_shape(self, px, py):
        # Updates the position of the active shape during drags.
        x0, y0 = self.__drag_start_pos
        dx = px - x0
        dy = py - y0
        self.__active_shape[:] = []
        for col in self.__active_shape_predrag:
            cx, cy = self.get_pos_for_color(col)
            cx += dx
            cy += dy
            col2 = super(HCYMaskEditorWheel, self).get_color_at_position(
                cx, cy, ignore_mask=True)
            self.__active_shape.append(col2)

    def __drag_active_ctrlpoint(self, px, py):
        # Moves the highlighted control point during drags.
        x0, y0 = self.__drag_start_pos
        dx = px - x0
        dy = py - y0
        col = self.__active_ctrlpoint_predrag
        cx, cy = self.get_pos_for_color(col)
        cx += dx
        cy += dy
        col = super(HCYMaskEditorWheel, self).get_color_at_position(
            cx, cy, ignore_mask=True)
        self.__active_shape[self.__active_ctrlpoint] = col

    def __rotate_mask(self, px, py):
        # Rotates the entire mask around the grey axis during drags.
        cx, cy = self.get_center()
        x0, y0 = self.__drag_start_pos
        theta0 = math.atan2(x0-cx, y0-cy)
        theta = math.atan2(px-cx, py-cy)
        dntheta = (theta0 - theta) / (2*math.pi)
        while dntheta <= 0:
            dntheta += 1.0
        if self.__mask_predrag is None:
            self.__mask_predrag = []
            for shape in self.get_mask():
                shape_hcy = [HCYColor(color=c) for c in shape]
                self.__mask_predrag.append(shape_hcy)
        mgr = self.get_color_manager()
        newmask = []
        for shape in self.__mask_predrag:
            shape_rot = []
            for col in shape:
                col_r = HCYColor(color=col)
                h = mgr.distort_hue(col_r.h)
                h += dntheta
                h %= 1.0
                col_r.h = mgr.undistort_hue(h)
                shape_rot.append(col_r)
            newmask.append(shape_rot)
        self.set_mask(newmask)

    def __button_press_cb(self, widget, event):
        # Begins drags.
        if self.__drag_func is None:
            self.__update_active_objects(event.x, event.y)
            self.__drag_start_pos = event.x, event.y
            if self.__tmp_new_ctrlpoint is not None:
                self.__active_ctrlpoint = len(self.__active_shape)
                self.__active_shape.append(self.__tmp_new_ctrlpoint)
                self.__tmp_new_ctrlpoint = None
            if self.__active_ctrlpoint is not None:
                self.__active_shape_predrag = self.__active_shape[:]
                ctrlpt = self.__active_shape[self.__active_ctrlpoint]
                self.__active_ctrlpoint_predrag = ctrlpt
                self.__drag_func = self.__drag_active_ctrlpoint
                self.__set_cursor(self.__move_point_cursor)
            elif self.__active_shape is not None:
                self.__active_shape_predrag = self.__active_shape[:]
                self.__drag_func = self.__drag_active_shape
                self.__set_cursor(self.__move_cursor)
            elif self.__last_cursor is self.__rotate_cursor:
                self.__mask_predrag = None
                self.__drag_func = self.__rotate_mask

    def __button_release_cb(self, widget, event):
        # Ends the current drag & cleans up, or handle other clicks.
        if self.__drag_func is None:
            # Clicking when not in a drag adds a new shape
            if self.__last_cursor is self.__add_cursor:
                self.__add_void(event.x, event.y)
        else:
            # Cleanup when dragging ends
            self.__drag_func = None
            self.__drag_start_pos = None
            self.__cleanup_mask()
        self.__update_active_objects(event.x, event.y)

    def __motion_cb(self, widget, event):
        # Fire the current drag function if one's active.
        if self.__drag_func is not None:
            self.__drag_func(event.x, event.y)
            self.queue_draw()
        else:
            self.__update_active_objects(event.x, event.y)

    def __cleanup_mask(self):
        mask = self.get_mask()

        # Drop points from all shapes which are not part of the convex hulls.
        for shape in mask:
            if len(shape) <= 3:
                continue
            points = [self.get_pos_for_color(c) for c in shape]
            edge_points = geom.convex_hull(points)
            for col, point in zip(shape, points):
                if point in edge_points:
                    continue
                shape.remove(col)

        # Drop shapes smaller than the minimum size.
        newmask = []
        min_size = self.get_radius() * self.min_shape_size
        for shape in mask:
            points = [self.get_pos_for_color(c) for c in shape]
            void = geom.convex_hull(points)
            size = self._get_void_size(void)
            if size >= min_size:
                newmask.append(shape)
        mask = newmask

        # Drop shapes whose points entirely lie within other shapes
        newmask = []
        maskvoids = [(shape, geom.convex_hull([self.get_pos_for_color(c)
                                               for c in shape]))
                     for shape in mask]
        for shape1, void1 in maskvoids:
            shape1_subsumed = True
            for p1 in void1:
                p1_subsumed = False
                for shape2, void2 in maskvoids:
                    if shape1 is shape2:
                        continue
                    if geom.point_in_convex_poly(p1, void2):
                        p1_subsumed = True
                        break
                if not p1_subsumed:
                    shape1_subsumed = False
                    break
            if not shape1_subsumed:
                newmask.append(shape1)
        mask = newmask

        self.set_mask(mask)
        self.queue_draw()

    def __dist_to_nearest_shape(self, x, y):
        # Distance from `x`, `y` to the nearest edge or vertex of any shape.
        dists = []
        for hull in self.get_mask_voids():
            # cx, cy = geom.poly_centroid(hull)
            for p1, p2 in geom.pairwise(hull):
                nearest_point = geom.nearest_point_in_segment(p1, p2, (x, y))
                if nearest_point is not None:
                    nx, ny = nearest_point
                    d = math.sqrt((x-nx)**2 + (y-ny)**2)
                    dists.append(d)
            # Segment end too
            d = math.sqrt((p1[0]-x)**2 + (p1[1]-y)**2)
            dists.append(d)
        if not dists:
            return None
        dists.sort()
        return dists[0]

    def __add_void(self, x, y):
        # Adds a new shape into the empty space centred at `x`, `y`.
        self.queue_draw()
        # Pick a nice size for the new shape, taking care not to
        # overlap any other shapes, at least initially.
        alloc = self.get_allocation()
        cx, cy = self.get_center(alloc=alloc)
        radius = self.get_radius(alloc=alloc)
        dx, dy = x-cx, y-cy
        r = math.sqrt(dx**2 + dy**2)
        d = self.__dist_to_nearest_shape(x, y)
        if d is None:
            d = radius
        size = min((radius - r), d) * 0.95
        minsize = radius * self.min_shape_size
        if size < minsize:
            return
        # Create a regular polygon with one of its edges facing the
        # middle of the wheel.
        shape = []
        nsides = 3 + len(self.get_mask())
        psi = math.atan2(dy, dx) + (math.pi/nsides)
        psi += math.pi
        for i in xrange(nsides):
            theta = 2.0 * math.pi * i / nsides
            theta += psi
            px = int(x + size*math.cos(theta))
            py = int(y + size*math.sin(theta))
            col = self.get_color_at_position(px, py, ignore_mask=True)
            shape.append(col)
        mask = self.get_mask()
        mask.append(shape)
        self.set_mask(mask)

    def draw_mask_control_points(self, cr, wd, ht):
        # Draw active and inactive control points on the active shape.

        if self.__active_shape is None:
            return

        cr.save()
        active_rgb = 1, 1, 1
        normal_rgb = 0, 0, 0
        delete_rgb = 1, 0, 0
        cr.set_line_width(1.0)
        void = self.colors_to_mask_void(self.__active_shape)

        # Highlight the objects that would be directly or indirectly affected
        # if the shape were dragged, and how.
        min_size = self.get_radius(wd=wd, ht=ht) * self.min_shape_size
        void_rgb = normal_rgb
        if self._get_void_size(void) < min_size:
            # Shape will be deleted
            void_rgb = delete_rgb
        elif ((self.__active_ctrlpoint is None) and
              (self.__tmp_new_ctrlpoint is None)):
            # The entire shape would be moved
            void_rgb = active_rgb
        # Outline the current shape
        cr.set_source_rgb(*void_rgb)
        for p_idx, p in enumerate(void):
            if p_idx == 0:
                cr.move_to(*p)
            else:
                cr.line_to(*p)
        cr.close_path()
        cr.stroke()

        # Control points
        colors = self.__active_shape
        for col_idx, col in enumerate(colors):
            px, py = self.get_pos_for_color(col)
            if (px, py) not in void:
                # not in convex hull (is it worth doing this fragile test?)
                continue
            point_rgb = void_rgb
            if col_idx == self.__active_ctrlpoint:
                point_rgb = active_rgb
            cr.set_source_rgb(*point_rgb)
            cr.arc(px, py, self.__ctrlpoint_radius, 0, 2*math.pi)
            cr.fill()
        if self.__tmp_new_ctrlpoint:
            px, py = self.get_pos_for_color(self.__tmp_new_ctrlpoint)
            cr.set_source_rgb(*active_rgb)
            cr.arc(px, py, self.__ctrlpoint_radius, 0, 2*math.pi)
            cr.fill()

        # Centroid
        cr.set_source_rgb(*void_rgb)
        cx, cy = geom.poly_centroid(void)
        cr.save()
        cr.set_line_cap(cairo.LINE_CAP_SQUARE)
        cr.set_line_width(0.5)
        cr.translate(int(cx)+0.5, int(cy)+0.5)
        cr.move_to(-2, 0)
        cr.line_to(2, 0)
        cr.stroke()
        cr.move_to(0, -2)
        cr.line_to(0, 2)
        cr.stroke()

        cr.restore()

    def paint_foreground_cb(self, cr, wd, ht):
        """Foreground drawing override.
        """
        self.draw_mask(cr, wd, ht)
        self.draw_mask_control_points(cr, wd, ht)

    def get_managed_color(self):
        """Override, with a limited range or returned luma.
        """
        col = super(HCYMaskEditorWheel, self).get_managed_color()
        col = HCYColor(color=col)
        col.y = clamp(col.y, self.__MIN_LUMA, self.__MAX_LUMA)
        return col

    def set_managed_color(self, color):
        """Override, limiting the luma range.
        """
        col = HCYColor(color=color)
        col.y = clamp(col.y, self.__MIN_LUMA, self.__MAX_LUMA)
        super(HCYMaskEditorWheel, self).set_managed_color(col)


class HCYMaskPreview (MaskableWheelMixin,
                      HCYHueChromaWheelMixin,
                      HueSaturationWheelAdjuster):
    """Mask preview widget; not scrollable.

    These widgets can be used with `paletteview.palette_load_via_dialog()` as
    preview widgets during mask selection.

    """

    def __init__(self):
        MaskableWheelMixin.__init__(self)
        HueSaturationWheelAdjuster.__init__(self)
        self.set_app_paintable(True)
        self.set_has_window(False)
        self.set_mask([])
        self.mask_toggle.set_active(True)
        self.set_size_request(64, 64)

    def render_background_cb(self, cr, wd, ht):
        sup = HueSaturationWheelAdjuster
        sup.render_background_cb(self, cr, wd=wd, ht=ht)
        self.draw_mask(cr, wd=wd, ht=ht)

    def paint_foreground_cb(self, cr, wd, ht):
        pass

    def get_background_validity(self):
        return deepcopy(self.get_mask())

    def set_palette(self, palette):
        # Compatibility with palette_load_via_dialog()
        self.set_mask_from_palette(palette)


class HCYMaskTemplateDialog (Gtk.Dialog):
    """Dialog for choosing a mask from a small set of templates.

    http://gurneyjourney.blogspot.co.uk/2008/02/shapes-of-color-schemes.html

    """

    @property
    def __templates(self):
        Y = 0.5
        H = 1-0.05
        # Reusable shapes...
        atmos_triad = [
            (H, 0.95, Y),
            ((H+0.275) % 1, 0.55, Y),
            ((1+H-0.275) % 1, 0.55, Y)
        ]

        def __coffin(h):
            # Hexagonal coffin shape with the foot end at the centre
            # of the wheel.
            shape = []
            shape.append(((h + 0.25) % 1, 0.03, Y))
            shape.append(((h + 1 - 0.25) % 1, 0.03, Y))
            shape.append(((h + 0.01) % 1, 0.95, Y))
            shape.append(((h + 1 - 0.01) % 1, 0.95, Y))
            shape.append(((h + 0.04) % 1, 0.70, Y))
            shape.append(((h + 1 - 0.04) % 1, 0.70, Y))
            return shape

        def __complement_blob(h):
            # Small pentagonal blob at the given hue, used for an organic-
            # looking dab of a complementary hue.
            shape = []
            shape.append(((h+0.015) % 1, 0.94, Y))
            shape.append(((h+0.985) % 1, 0.94, Y))
            shape.append(((h+0.035) % 1, 0.71, Y))
            shape.append(((h+0.965) % 1, 0.71, Y))
            shape.append(((h) % 1, 0.54, Y))
            return shape

        templates = []
        templates.append((
            C_(
                "HCY Gamut Mask template name",
                u"Atmospheric Triad",
            ),
            C_(
                "HCY Gamut Mask template description",
                "Moody and subjective, defined by one dominant primary "
                "and two primaries which are less intense.",
            ),
            [deepcopy(atmos_triad)]
        ))
        templates.append((
            C_(
                "HCY Gamut Mask template name",
                u"Shifted Triad",
            ),
            C_(
                "HCY Gamut Mask template description",
                u"Weighted more strongly towards the dominant color.",
            ),
            [
                [(H, 0.95, Y),
                 ((H+0.35) % 1, 0.4, Y),
                 ((1+H-0.35) % 1, 0.4, Y)]
            ]
        ))
        templates.append((
            C_(
                "HCY Gamut Mask template name",
                u"Complementary",
            ),
            C_(
                "HCY Gamut Mask template description",
                u"Contrasting opposites, "
                u"balanced by having central neutrals "
                u"between them on the color wheel.",
            ),
            [
                [((H+0.005) % 1, 0.9, Y),
                 ((H+0.995) % 1, 0.9, Y),
                 ((H+0.250) % 1, 0.1, Y),
                 ((H+0.750) % 1, 0.1, Y),
                 ((H+0.505) % 1, 0.9, Y),
                 ((H+0.495) % 1, 0.9, Y)]
            ]
        ))
        templates.append((
            C_(
                "HCY Gamut Mask template name",
                u"Mood and Accent",
            ),
            C_(
                "HCY Gamut Mask template description",
                u"One main range of colors, "
                u"with a complementary accent for "
                u"variation and highlights.",
            ),
            [deepcopy(atmos_triad), __complement_blob(H+0.5)]
        ))
        templates.append((
            C_(
                "HCY Gamut Mask template name",
                u"Split Complementary",
            ),
            C_(
                "HCY Gamut Mask template description",
                u"Two analogous colors and a complement to them, "
                u"with no secondary colors between them.",
            ),
            [__coffin(H+0.5), __coffin(1+H-0.1), __coffin(H+0.1)]
        ))
        return templates

    def __init__(self, parent, target):
        super(HCYMaskTemplateDialog, self).__init__(
            title=C_(
                u"HCY Gamut Mask new-from-template dialog: window title",
                "New Gamut Mask from Template",
            ),
            transient_for=parent,
            modal=True,
            destroy_with_parent=True,
            window_position=Gtk.WindowPosition.MOUSE,
        )
        self.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.REJECT)
        target_mgr = target.get_color_manager()
        prefs_ro = deepcopy(target_mgr.get_prefs())
        datapath = target_mgr.get_data_path()
        mgr = ColorManager(prefs=prefs_ro, datapath=datapath)
        mgr.set_wheel_type(target_mgr.get_wheel_type())
        self.target = target
        for name, desc, mask_shapes_float in self.__templates:
            mask = []
            for mask_shape_float in mask_shapes_float:
                shape = []
                for h, c, y in mask_shape_float:
                    h = mgr.undistort_hue(h)
                    shape.append(HCYColor(h, c, y))
                mask.append(shape)
            label = Gtk.Label()
            label.set_markup("<b>%s</b>\n\n%s" % (name, desc))
            label.set_size_request(375, -1)
            label.set_line_wrap(True)
            label.set_alignment(0, 0.5)
            preview = HCYMaskPreview()
            preview.set_color_manager(mgr)
            preview.set_mask(mask)
            preview_frame = Gtk.AspectFrame(obey_child=True)
            preview_frame.add(preview)
            preview_frame.set_shadow_type(Gtk.ShadowType.NONE)
            hbox = Gtk.HBox()
            hbox.set_spacing(6)
            hbox.pack_start(preview_frame, False, False, 0)
            hbox.pack_start(label, True, True, 0)
            button = Gtk.Button()
            button.add(hbox)
            button.set_relief(Gtk.ReliefStyle.NONE)
            button.connect("clicked", self.__button_clicked_cb, mask)
            self.vbox.pack_start(button, True, True, 0)
        self.connect("response", self.__response_cb)
        self.connect("show", self.__show_cb)
        for w in self.vbox:
            w.show_all()
        ref_color = target.get_managed_color()
        mgr.set_color(ref_color)

    def __button_clicked_cb(self, widget, mask):
        self.target.set_mask(mask)
        self.hide()

    def __show_cb(self, widget, *a):
        self.vbox.show_all()

    def __response_cb(self, widget, response_id):
        self.hide()
        return True


class HCYMaskPropertiesDialog (Gtk.Dialog):
    """Dialog for choosing, editing, or enabling/disabling masks.
    """

    def __init__(self, parent, target):
        super(HCYMaskPropertiesDialog, self).__init__(
            title=C_(
                "HCY Gamut Mask Editor dialog: window title",
                u"Gamut Mask Editor",
            ),
            transient_for=parent,
            modal=True,
            destroy_with_parent=True,
            window_position=Gtk.WindowPosition.CENTER_ON_PARENT,
            buttons=(
                Gtk.STOCK_CANCEL, Gtk.ResponseType.REJECT,
                Gtk.STOCK_OK, Gtk.ResponseType.ACCEPT,
            ),
        )
        self.target = target
        ed = HCYMaskEditorWheel()
        target_mgr = target.get_color_manager()
        prefs_ro = deepcopy(target_mgr.get_prefs())
        datapath = target_mgr.get_data_path()
        ed_mgr = ColorManager(prefs=prefs_ro, datapath=datapath)
        ed.set_color_manager(ed_mgr)
        self.editor = ed
        ed.set_size_request(300, 300)
        ed.mask_toggle.set_active(True)
        self.mask_toggle_ctrl = Gtk.CheckButton(
            C_(
                "HCY Gamut Mask Editor dialog: mask-is-active checkbox",
                u"Active",
            ),
            use_underline=False,
        )
        self.mask_toggle_ctrl.set_tooltip_text(ed.mask_toggle.get_tooltip())
        ed.mask_observers.append(self.__mask_changed_cb)

        hbox = Gtk.HBox()
        hbox.set_spacing(3)

        # Sidebar buttonbox
        # On the right and packed to the top. This places its secondary
        # control, a mask toggle button, next to the "OK" button so it's less
        # likely to be missed.
        bbox = Gtk.VButtonBox()
        new_btn = self.__new_button = Gtk.Button(stock=Gtk.STOCK_NEW)
        load_btn = self.__load_button = Gtk.Button(stock=Gtk.STOCK_OPEN)
        save_btn = self.__save_button = Gtk.Button(stock=Gtk.STOCK_SAVE)
        clear_btn = self.__clear_button = Gtk.Button(stock=Gtk.STOCK_CLEAR)

        help_btn = self.__help_button = Gtk.LinkButton.new_with_label(
            uri = MASK_EDITOR_HELP_URI,
            label = C_(
                "HCY Mask Editor: action button labels",
                u"Help…",
            ),
        )

        new_btn.set_tooltip_text(C_(
            "HCY Mask Editor: action button tooltips",
            u"Create mask from template."),
        )
        load_btn.set_tooltip_text(C_(
            "HCY Mask Editor: action button tooltips",
            u"Load mask from a GIMP palette file."),
        )
        save_btn.set_tooltip_text(C_(
            "HCY Mask Editor: action button tooltips",
            u"Save mask to a GIMP palette file."),
        )
        clear_btn.set_tooltip_text(C_(
            "HCY Mask Editor: action button tooltips",
            u"Erase the mask."),
        )
        help_btn.set_tooltip_text(C_(
            "HCY Mask Editor: action button tooltips",
            u"Open the online help for this dialog in a web browser."),
        )

        new_btn.connect("clicked", self.__new_clicked)
        save_btn.connect("clicked", self.__save_clicked)
        load_btn.connect("clicked", self.__load_clicked)
        clear_btn.connect("clicked", self.__clear_clicked)

        bbox.pack_start(new_btn, True, True, 0)
        bbox.pack_start(load_btn, True, True, 0)
        bbox.pack_start(save_btn, True, True, 0)
        bbox.pack_start(clear_btn, True, True, 0)

        action_area = self.get_action_area()
        if isinstance(action_area, Gtk.ButtonBox):
            action_area.pack_start(help_btn, True, True, 0)
            action_area.set_child_secondary(help_btn, True)
            action_area.set_child_non_homogeneous(help_btn, True)
            bbox.pack_start(self.mask_toggle_ctrl, True, True, 0)
            bbox.set_child_secondary(self.mask_toggle_ctrl, True)
        else:
            bbox.pack_start(self.mask_toggle_ctrl, True, True, 0)
            bbox.pack_start(help_btn, True, True, 0)
            bbox.set_child_secondary(help_btn, True)

        bbox.set_layout(Gtk.ButtonBoxStyle.START)

        hbox.pack_start(ed, True, True, 0)
        hbox.pack_start(bbox, False, False, 0)
        hbox.set_border_width(9)

        self.vbox.pack_start(hbox, True, True, 0)

        self.connect("response", self.__response_cb)
        self.connect("show", self.__show_cb)
        for w in self.vbox:
            w.show_all()

    def __mask_changed_cb(self):
        mask = self.editor.get_mask()
        empty = mask == []
        self.__save_button.set_sensitive(not empty)
        self.__clear_button.set_sensitive(not empty)

    def __new_clicked(self, widget):
        dialog = HCYMaskTemplateDialog(self, self.editor)
        dialog.run()

    def __save_clicked(self, button):
        pal = Palette()
        mask = self.editor.get_mask()
        for i, shape in enumerate(mask):
            for j, col in enumerate(shape):
                col_name = "mask#%d primary#%d" % (i, j)  # NOT localised
                pal.append(col, col_name)
        preview = HCYMaskPreview()
        preview.set_size_request(128, 128)
        target_mgr = self.target.get_color_manager()
        prefs_ro = deepcopy(target_mgr.get_prefs())
        datapath = target_mgr.get_data_path()
        mgr = ColorManager(prefs=prefs_ro, datapath=datapath)
        preview.set_color_manager(mgr)
        preview.set_managed_color(self.editor.get_managed_color())
        palette_save_via_dialog(
            pal,
            title = C_(
                "HCY Gamut Mask load dialog: window title",
                u"Save Mask as a GIMP Palette"
            ),
            parent = self,
            preview = preview,
        )

    def __load_clicked(self, button):
        preview = HCYMaskPreview()
        preview.set_size_request(128, 128)
        target_mgr = self.target.get_color_manager()
        prefs_ro = deepcopy(target_mgr.get_prefs())
        datapath = target_mgr.get_data_path()
        mgr = ColorManager(prefs=prefs_ro, datapath=datapath)
        preview.set_color_manager(mgr)
        preview.set_managed_color(self.editor.get_managed_color())
        dialog_title = C_(
            "HCY Gamut Mask load dialog: window title",
            u"Load Mask from a GIMP Palette",
        )
        pal = palette_load_via_dialog(title=dialog_title, parent=self,
                                      preview=preview)
        if pal is None:
            return
        self.editor.set_mask_from_palette(pal)

    def __clear_clicked(self, widget):
        self.editor.set_mask([])

    def __show_cb(self, widget, *a):
        # When the dialog is shown, clone the target adjuster's mask for
        # editing. Assume the user wants to turn on the mask if there
        # is no mask on the target already (reduce the number of mouse clicks)
        active = True
        if self.target.get_mask():
            active = self.target.mask_toggle.get_active()
        self.mask_toggle_ctrl.set_active(active)
        mask = deepcopy(self.target.get_mask())
        self.editor.set_mask(mask)

        # The wheel type may have changed elsewhere
        editor_mgr = self.editor.get_color_manager()
        wheel_type = self.target.get_color_manager().get_wheel_type()
        editor_mgr.set_wheel_type(wheel_type)

        # Clone the target's luma too,
        # but not too bright, not too dark
        col = HCYColor(color=self.target.get_managed_color())
        self.editor.set_managed_color(col)

        # Necessary for the content to be displayed
        self.vbox.show_all()

    def __response_cb(self, widget, response_id):
        if response_id == Gtk.ResponseType.ACCEPT:
            self.target.set_mask(self.editor.get_mask())
            mask_active = self.mask_toggle_ctrl.get_active()
            self.target.mask_toggle.set_active(mask_active)
        self.hide()


class HCYAdjusterPage (CombinedAdjusterPage):
    """Combined HCY adjuster.
    """

    def __init__(self):
        y_adj = HCYLumaSlider()
        y_adj.vertical = True
        hc_adj = HCYHueChromaWheel()

        table = Gtk.Table(rows=2, columns=2)
        xopts = Gtk.AttachOptions.FILL | Gtk.AttachOptions.EXPAND
        yopts = Gtk.AttachOptions.FILL | Gtk.AttachOptions.EXPAND
        table.attach(y_adj, 0, 1, 0, 1, Gtk.AttachOptions.FILL, yopts, 3, 3)
        table.attach(hc_adj, 1, 2, 0, 2, xopts, yopts, 3, 3)

        self.__y_adj = y_adj
        self.__hc_adj = hc_adj
        self.__table = table
        self.__mask_dialog = None

    @classmethod
    def get_properties_description(cls):
        return C_(
            "HCY Wheel color adjuster page: properties tooltip.",
            u"Set gamut mask.",
        )

    def show_properties(self):
        if self.__mask_dialog is None:
            toplevel = self.__hc_adj.get_toplevel()
            dia = HCYMaskPropertiesDialog(toplevel, self.__hc_adj)
            self.__mask_dialog = dia
        self.__mask_dialog.run()

    @classmethod
    def get_page_icon_name(cls):
        return 'mypaint-tool-hcywheel'

    @classmethod
    def get_page_title(cls):
        return C_(
            "HCY Wheel color adjuster page: title for tooltips etc.",
            u"HCY Wheel",
        )

    @classmethod
    def get_page_description(cls):
        return C_(
            "HCY Wheel color adjuster page: description for tooltips etc.",
            u"Set the color using cylindrical hue/chroma/luma space. "
            u"The circular slices are equiluminant.",
        )

    def get_page_widget(self):
        frame = Gtk.AspectFrame(obey_child=True)
        frame.set_shadow_type(Gtk.ShadowType.NONE)
        frame.add(self.__table)
        return frame

    def set_color_manager(self, manager):
        ColorAdjuster.set_color_manager(self, manager)
        self.__y_adj.set_property("color-manager", manager)
        self.__hc_adj.set_property("color-manager", manager)


if __name__ == '__main__':
    import os
    import sys
    mgr = ColorManager(prefs={}, datapath='.')
    mgr.set_color(HSVColor(0.0, 0.0, 0.55))
    if len(sys.argv) > 1:
        # Generate icons
        wheel = HCYHueChromaWheel()
        wheel.set_color_manager(mgr)
        icon_name = HCYAdjusterPage.get_page_icon_name()
        for dir_name in sys.argv[1:]:
            wheel.save_icon_tree(dir_name, icon_name)
    else:
        # Interactive test
        page = HCYAdjusterPage()
        page.set_color_manager(mgr)
        window = Gtk.Window()
        window.add(page.get_page_widget())
        window.set_title(os.path.basename(sys.argv[0]))
        window.set_border_width(6)
        window.connect("destroy", lambda *a: Gtk.main_quit())
        window.show_all()
        Gtk.main()



# This file is part of MyPaint.
# Copyright (C) 2007-2013 by Martin Renold <martinxyz@gmx.ch>
# Copyright (C) 2013-2018 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""Command-line handling - traditional main() function."""

## Imports *nothing involving mypaintlib at this point*

from __future__ import division, print_function

import os
import sys
import logging
import warnings

from lib.gibindings import GdkPixbuf
from optparse import OptionParser

import lib.config
import lib.glib
from lib.i18n import USER_LOCALE_PREF
from lib.meta import MYPAINT_VERSION
import gui.userconfig

logger = logging.getLogger(__name__)


## Method defs

def _init_gtk_workarounds():
    """Initialize some workarounds for unoptimal GTK behavior"""
    # Via https://code.google.com/p/quodlibet/source/browse/quodlibet/
    logger.debug("Adding GTK workarounds...")

    # On windows the default variants only do ANSI paths, so replace them.
    # In some typelibs they are replaced by default, in some don't..
    if os.name == "nt":
        for name in ["new_from_file_at_scale", "new_from_file_at_size",
                     "new_from_file"]:
            cls = GdkPixbuf.Pixbuf
            func = getattr(cls, name + "_utf8", None)
            if func:
                logger.debug(
                    "Monkeypatching GdkPixbuf.Pixbuf.%s with %r",
                    name, func,
                )
                setattr(cls, name, func)

    # Wayland "workaround" to avoid input freeze on pointer grabs.
    # Respect existing envvars for testing (and general courtesy).
    # This relies on XWayland being available.
    if sys.platform.startswith("linux") and 'GDK_BACKEND' not in os.environ:
        os.environ['GDK_BACKEND'] = 'x11'

    logger.debug("GTK workarounds added.")


def set_user_configured_locale(userconfpath):
    """If configured, set envvars for a custom locale
    The user can choose to not use their system locale
    by explicitly choosing another, and restarting.
    """
    settings = gui.userconfig.get_json_config(userconfpath)
    if USER_LOCALE_PREF in settings:
        user_locale = settings[USER_LOCALE_PREF]
        supported = ["en", "en_US"] + lib.config.supported_locales
        if user_locale not in supported:
            logger.warning("Locale not supported: %s", user_locale)
        else:
            # GIMP, Krita and Inkscape will all ignore the LANGUAGE
            # variable when a user has chosen a language explicitly.
            # For better or worse, we follow their example
            # (this may change in the future).
            logger.info("Using locale: (%s)", user_locale)
            if "LANGUAGE" in os.environ:
                logger.warning("LANGUAGE envvar is overridden by user setting")
            os.environ["LANGUAGE"] = user_locale
    else:
        logger.info("No locale setting found, using system locale")


def main(
        datapath,
        iconspath,
        localepath,
        oldstyle_confpath=None,
        version=MYPAINT_VERSION,
        debug=False
):
    """Run MyPaint with `sys.argv_unicode`, called from the "mypaint" script.

    :param unicode datapath: The app's read-only data location.
    :param unicode iconspath: Extra search root for finding icons.
    :param unicode oldstyle_confpath: Old-style merged config folder.
    :param unicode version: full version string for the about box.
    :param bool debug: whether debug functionality and logging is enabled.

    The ``datapath`` parameter defines where MyPaint should find its
    static data, e.g. UI definition XML, backgrounds, and brush
    definitions. $PREFIX/share/mypaint is usual.

    The iconspath`` parameter tells MyPaint where to find its themeable UI
    icons. This will be used in addition to $XDG_DATA_DIRS for the
    purposes of icon lookup.  Normally it's $PREFIX/share/icons.

    If specified oldstyle_confpath must be a single directory path.  all
    user-specific data that MyPaint writes is written here.  If omitted,
    this data will be stored under the basedirs returned by
    GLib.get_user_config_dir() for settings, and by
    GLib.get_user_data_dir() for brushes and backgrounds. On Windows,
    these will be the same location.  On POSIX systems,
    $HOME/.config/mypaint and $HOME/.local/share/mypaint are a typical
    division.

    See `lib.meta` for details of what normally goes in `version`.

    """

    # Init the workaround before we start importing anything which
    # could still be using gtk2compat.
    _init_gtk_workarounds()

    # GLib user dirs: cache them now for greatest compatibility.
    # Importing mypaintlib before the 1st call to g_get_user*_dir()
    # breaks GLib for obscure reasons.
    # This needs to be done after i18n setup, or Windows configurations
    # with non-ASCII character in %USERPROFILE% will break.
    lib.glib.init_user_dir_caches()

    options, args = parsed_cmdline_arguments(oldstyle_confpath, debug)

    # XDG support for new users on POSIX platforms
    if options.config is None:
        appsubdir = u"mypaint"
        basedir = lib.glib.get_user_data_dir()
        userdatapath = os.path.join(basedir, appsubdir)
        basedir = lib.glib.get_user_config_dir()
        userconfpath = os.path.join(basedir, appsubdir)
    else:
        userdatapath = options.config
        userconfpath = options.config

    # Log to the specified location
    # Prepend the data path if the user gave a relative location.
    if options.logfile:
        logfilepath = os.path.join(userdatapath, options.logfile)
        logdirpath, logfilebasename = os.path.split(logfilepath)
        if not os.path.isdir(logdirpath):
            os.makedirs(logdirpath)
        logger.info("Copying log messages to %r", logfilepath)
        logfile_handler = logging.FileHandler(
            logfilepath, mode="a",
            encoding="utf-8",
        )
        logfile_format = "%(asctime)s;%(levelname)s;%(name)s;%(message)s"
        logfile_handler.setFormatter(logging.Formatter(logfile_format))
        root_logger = logging.getLogger(None)
        root_logger.addHandler(logfile_handler)

    if debug:
        logger.critical("Test critical message, please ignore")
        warnings.resetwarnings()
        logging.captureWarnings(True)

    if options.version:
        # Output (rather than log) the version
        print("MyPaint version %s" % (version, ))
        sys.exit(0)

    def run():
        logger.debug('user_datapath: %r', userdatapath)
        logger.debug('user_confpath: %r', userconfpath)

        # User-configured locale (if enabled by user)
        set_user_configured_locale(userconfpath)

        # Locale setting
        from lib.gettext_setup import init_gettext
        init_gettext(localepath)

        # mypaintlib import is performed first in gui.application now.
        from gui import application

        app_state_dirs = application.StateDirs(
            app_data = datapath,
            app_icons = iconspath,
            user_data = userdatapath,
            user_config = userconfpath,
        )
        app = application.Application(
            filenames = args,
            state_dirs = app_state_dirs,
            version = version,
            fullscreen = options.fullscreen,
        )

        # Gtk must not be imported before init_gettext
        # has been run - else locales will not be set
        # up properly (e.g: left-to-right interfaces for right-to-left scripts)
        # Note that this is not the first import of Gtk in the __program__;
        # it is imported indirectly via the import of gui.application
        from lib.gibindings import Gtk
        settings = Gtk.Settings.get_default()
        dark = app.preferences.get("ui.dark_theme_variant", True)
        settings.set_property("gtk-application-prefer-dark-theme", dark)

        if debug and options.run_and_quit:
            from lib.gibindings import GLib
            GLib.timeout_add(1000, lambda *a: Gtk.main_quit())
        else:
            from gui import gtkexcepthook
            func = app.filehandler.confirm_destructive_action
            gtkexcepthook.quit_confirmation_func = func

        # temporary workaround for gtk3 Ctrl-C bug:
        # https://bugzilla.gnome.org/show_bug.cgi?id=622084
        import signal
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        Gtk.main()

    if options.trace:
        import trace
        tracer = trace.Trace(trace=1, count=0)
        tracer.runfunc(run)
    else:
        run()


def parsed_cmdline_arguments(default_confpath, debug=False):
    """Parse command line arguments and return result

    :return: (options, positional arguments)
    """
    parser = OptionParser('usage: %prog [options] [FILE]')
    parser.add_option(
        '-c',
        '--config',
        metavar='DIR',
        default=default_confpath,
        help='use old-style merged config directory DIR, e.g. ~/.mypaint'
    )
    parser.add_option(
        '-l',
        '--logfile',
        metavar='FILE',
        default=None,
        help='log console messages to FILE (rel. to config location)'
    )
    parser.add_option(
        '-t',
        '--trace',
        action="store_true",
        help='print all executed Python statements'
    )
    parser.add_option(
        '-f',
        '--fullscreen',
        action="store_true",
        help='start in fullscreen mode'
    )
    parser.add_option(
        "-V",
        '--version',
        action="store_true",
        help='print version information and exit'
    )
    if debug:
        parser.add_option(
            "-R",
            '--run-and-quit',
            action="store_true",
            help='start the program and shut it down after 1 second'
        )

    return parser.parse_args(sys.argv_unicode[1:])



# This file is part of MyPaint.
# Copyright (C) 2018-2019 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""This module implements tile-based floodfill and related operations."""

import logging

import numpy as np
import threading

from lib.gibindings import GLib

import lib.helpers
import lib.mypaintlib as myplib
import lib.surface
import lib.tiledsurface
from lib.gettext import C_
import lib.fill_common as fc
from lib.fill_common import _OPAQUE, _FULL_TILE, _EMPTY_TILE
import lib.modes
import lib.morphology

from lib.pycompat import iteritems

logger = logging.getLogger(__name__)

TILE_SIZE = N = myplib.TILE_SIZE
INF_DIST = 2*N*N

# This should point to the array transparent_tile.rgba
# defined in tiledsurface.py
_EMPTY_RGBA = None

# Distance data for tiles with no detected distances
_GAPLESS_TILE = fc.new_full_tile(INF_DIST)
_GAPLESS_TILE.flags.writeable = False

EDGE = myplib.edges


class GapClosingOptions:
    """Container of parameters for gap closing fill operations
    to avoid updates to the call chain in case the parameter set
    is altered.
    """
    def __init__(self, max_gap_size, retract_seeps):
        self.max_gap_size = max_gap_size
        self.retract_seeps = retract_seeps


def enqueue_overflows(queue, tile_coord, seeds, tiles_bbox, *p):
    """ Conditionally add (coordinate, seed list, data...) tuples to a queue.

    :param queue: the queue which may be appended
    :type queue: list
    :param tile_coord: the 2d coordinate in the middle of the seed coordinates
    :type tile_coord: (int, int)
    :param seeds: 4-tuple of seed lists for n, e, s, w, relative to tile_coord
    :type seeds: (list, list, list, list)
    :param tiles_bbox: the bounding box of the fill operation
    :type tiles_bbox: lib.fill_common.TileBoundingBox
    :param p: tuples of length >= 4, items added to queue items w. same index

    NOTE: This function improves readability significantly in exchange for a
    small performance hit. Replace with explicit queueing if too slow.
    """
    for edge in zip(*(fc.orthogonal(tile_coord), seeds) + p):
        edge_coord = edge[0]
        edge_seeds = edge[1]
        if edge_seeds and not tiles_bbox.outside(edge_coord):
            queue.append(edge)


def starting_coordinates(x, y):
    """Get the coordinates of starting tile and pixel (tx, ty, px, py)"""
    init_tx, init_ty = int(x // N), int(y // N)
    init_x, init_y = int(x % N), int(y % N)
    return init_tx, init_ty, init_x, init_y


def seeds_by_tile(seeds):
    """Partition and convert seed coordinates

    Partition a list of model-space seed coordinates into lists of
    in-tile coordinates associated to their respective tile coordinate
    in a dictionary.
    """
    tile_seeds = dict()
    for (x, y) in seeds:
        tx, ty, px, py = starting_coordinates(x, y)
        seed_list = tile_seeds.get((tx, ty), [])
        seed_list.append((px, py))
        tile_seeds[(tx, ty)] = seed_list
    return tile_seeds


def get_target_color(src, tx, ty, px, py):
    """Get the pixel color for the given tile/pixel coordinates"""
    with src.tile_request(tx, ty, readonly=True) as start:
        targ_r, targ_g, targ_b, targ_a = [
            int(c) for c in start[py][px]
        ]
    if targ_a == 0:
        targ_r, targ_g, targ_b = 0, 0, 0

    return targ_r, targ_g, targ_b, targ_a


# Main fill interface

class FillHandler:
    """Handles fill status and cancellation

    The fill is run in a separate thread, and this controller
    is used to start and (optionally) cancel the fill, as well
    as provide information about its current status.
    """

    # Stages that the fill operation can be in
    FILL = 0
    MORPH = 1
    BLUR = 2
    COMPOSITE = 3
    FINISHING = 4

    STAGE_STRINGS = [
        C_("floodfill status message: use active tense", "Filling"),
        C_("floodfill status message: use active tense", "Morphing"),
        C_("floodfill status message: use active tense", "Blurring"),
        C_("floodfill status message: use active tense", "Compositing"),
        C_("floodfill status message: use active tense", "Finishing up")
    ]
    TILES_STRING = C_("uniform square region of pixels, plural noun", "tiles")
    TILES_TEMPLATE = "{t} " + TILES_STRING

    def __init__(self):
        # An object with a "keep running" flag and "tiles processed"
        # data, for easier C++ access in morph/blur stages
        self.controller = myplib.Controller()
        # Separate "keep running" flag checked in Python code
        self.run = True
        self.stage = None
        self.set_stage(self.FILL)
        # When morphing, blurring or compositing,
        # this is the total amount of tiles to process
        self.tiles_max = 0
        self.fill_thread = None

    @property
    def tiles_processed(self):
        """The number of tiles processed for the current stage"""
        return self.controller.num_processed()

    def inc_processed(self):
        """Increment the number of tiles processed by 1"""
        self.controller.inc_processed(1)

    def set_stage(self, stage, num_tiles_to_process=None):
        """Change stage, updating strings and tile data"""
        self.stage = stage
        self.controller.reset()
        self.stage_string = self.STAGE_STRINGS[stage]
        if num_tiles_to_process:
            self.tiles_max = num_tiles_to_process
            self.stage_string += (
                " " + self.TILES_TEMPLATE.format(t=num_tiles_to_process))

    @property
    def progress_string(self):
        """Progress for the current stage"""
        if self.stage == self.FILL:
            return self.TILES_TEMPLATE.format(t=self.tiles_processed)
        elif self.stage < self.FINISHING:
            return str(int(100*self.tiles_processed/self.tiles_max)) + "%"
        else:
            return ""

    def wait(self, t=None):
        """Wait t seconds for the fill to complete"""
        self.fill_thread.join(t)

    def running(self):
        """Check if the fill is still running"""
        return self.fill_thread.is_alive()

    def cancel(self):
        """Tell the fill to stop as soon as possible"""
        self.controller.stop()
        self.run = False


class FloodFillArguments(object):
    """Container holding a set of flood fill arguments
    The purpose of this class is to avoid unnecessary
    call chain updates when changing/adding parameters.
    """

    def __init__(
            self, target_pos, seeds, color, tolerance, offset, feather,
            gap_closing_options, mode, lock_alpha, opacity, framed, bbox
    ):
        """ Create a new fill argument set
        :param target_pos: pixel coordinate of target color
        :type target_pos: tuple
        :param seeds: set of seed pixel coordinates {(x, y)...}
        :type seeds: set
        :param color: an RGB color
        :type color: tuple
        :param tolerance: how much filled pixels are permitted to vary
        :type tolerance: float [0.0, 1.0]
        :param offset: the post-fill expansion/contraction radius in pixels
        :type offset: int [-TILE_SIZE, TILE_SIZE]
        :param feather: the amount to blur the fill, after offset is applied
        :type feather: int [0, TILE_SIZE]
        :param gap_closing_options: parameters for gap closing fill, or None
        :type gap_closing_options: GapClosingOptions
        :param mode: Fill blend mode - normal, erasing or alpha locked
        :type mode: int (Any of the Combine* modes in mypaintlib)
        :param lock_alpha: Lock alpha of the destination layer
        :type lock_alpha: bool
        :param opacity: opacity of the fill
        :type opacity: float
        :param framed: Whether the frame is enabled or not.
        :type framed: bool
        :param bbox: Bounding box: limits the fill
        :type bbox: lib.helpers.Rect or equivalent 4-tuple
        """
        self.target_pos = target_pos
        self.seeds = seeds
        self.color = color
        self.tolerance = tolerance
        self.offset = offset
        self.feather = feather
        self.gap_closing_options = gap_closing_options
        self.mode = mode
        self.lock_alpha = lock_alpha
        self.opacity = opacity
        self.framed = framed
        self.bbox = bbox

    def skip_empty_dst(self):
        """If true, compositing to empty tiles does nothing"""
        return (
            self.lock_alpha or self.mode in [
                myplib.CombineSourceAtop,
                myplib.CombineDestinationOut,
                myplib.CombineDestinationIn,
            ]
        )

    def no_op(self):
        """If true, compositing will never alter the output layer

        These are comp modes for which alpha locking does not really
        make any sense, as any visible change caused by them requires
        the alpha of the destination to change as well.
        """
        return self.lock_alpha and (
            self.mode in lib.modes.MODES_DECREASING_BACKDROP_ALPHA
        )


def flood_fill(src, fill_args, dst):
    """Top-level fill interface
    Delegates actual fill in separate thread and returns a FillHandler

    :param src: source, surface-like object
    :type src: anything supporting readonly tile_request()
    :param fill_args: arguments common to all fill calls
    :type fill_args: FloodFillArguments
    :param dst: target surface
    :type dst: lib.tiledsurface.MyPaintSurface
    """
    handler = FillHandler()
    fill_function_args = (src, fill_args, dst, handler)
    fill_thread = threading.Thread(target=_flood_fill, args=fill_function_args)
    handler.fill_thread = fill_thread
    fill_thread.start()
    return handler


def _flood_fill(src, args, dst, handler):
    """Main flood fill function

    The fill is performed with reference to src.
    The resulting tiles are composited into dst.

    :param src: source, surface-like object
    :type src: anything supporting readonly tile_request()
    :param args: arguments common to all fill calls
    :type args: FloodFillArguments
    :param dst: target surface
    :type dst: lib.tiledsurface.MyPaintSurface
    :param handler: controller used to track state and cancel fill
    :type handler: FillHandler
    """
    _, _, width, height = args.bbox
    if width <= 0 or height <= 0 or args.no_op():
        return

    tiles_bbox = fc.TileBoundingBox(args.bbox)

    # Basic safety clamping
    tolerance = lib.helpers.clamp(args.tolerance, 0.0, 1.0)
    offset = lib.helpers.clamp(args.offset, -TILE_SIZE, TILE_SIZE)
    feather = lib.helpers.clamp(args.feather, 0, TILE_SIZE)

    # Initial parameters
    target_color = get_target_color(
        src, *starting_coordinates(*args.target_pos)
    )
    filler = myplib.Filler(*(target_color + (tolerance,)))
    seed_lists = seeds_by_tile(args.seeds)

    fill_args = (handler, src, seed_lists, tiles_bbox, filler)

    if args.gap_closing_options:
        fill_args += (args.gap_closing_options,)
        filled = gap_closing_fill(*fill_args)
    else:
        filled = scanline_fill(*fill_args)

    # Dilate/Erode (Grow/Shrink)
    if offset != 0 and handler.run:
        filled = lib.morphology.morph(handler, offset, filled)

    # Feather (Gaussian blur)
    if feather != 0 and handler.run:
        filled = lib.morphology.blur(handler, feather, filled)

    # When dilating or blurring the fill, only respect the
    # bounding box limits if they are set by an active frame
    trim_result = args.framed and (offset > 0 or feather != 0)
    if handler.run:
        composite(
            handler, args,
            trim_result, filled, tiles_bbox, dst
        )


def update_bbox(bbox, tx, ty):
    """Update given the min/max, x/y bounding box
    If a coordinate lies outside of the current
    bounds, set the bounds based on that coordinate
    """
    if bbox:
        min_tx, min_ty, max_tx, max_ty = bbox
        if tx < min_tx:
            min_tx = tx
        elif tx > max_tx:
            max_tx = tx
        if ty < min_ty:
            min_ty = ty
        elif ty > max_ty:
            max_ty = ty
        return min_tx, min_ty, max_tx, max_ty
    else:
        return tx, ty, tx, ty


def composite(
        handler, fill_args,
        trim_result, filled, tiles_bbox, dst):
    """Composite the filled tiles into the destination surface"""

    handler.set_stage(handler.COMPOSITE, len(filled))

    fill_col = fill_args.color

    # Prepare opaque color rgba tile for copying
    full_rgba = myplib.rgba_tile_from_alpha_tile(
        _FULL_TILE, *(fill_col + (0, 0, N-1, N-1)))

    # Bounding box of tiles that need updating
    dst_changed_bbox = None
    dst_tiles = dst.get_tiles()

    skip_empty_dst = fill_args.skip_empty_dst()
    mode = fill_args.mode
    lock_alpha = fill_args.lock_alpha
    opacity = fill_args.opacity

    tile_combine = myplib.tile_combine

    # Composite filled tiles into the destination surface
    for tile_coord, src_tile in iteritems(filled):

        if not handler.run:
            break

        handler.inc_processed()

        # Omit tiles outside of the bounding box _if_ the frame is enabled
        # Note:filled tiles outside bbox only originates from dilation/blur
        if trim_result and tiles_bbox.outside(tile_coord):
            continue

        # Skip empty destination tiles for erasing and alpha locking
        # Avoids completely unnecessary tile allocation and copying
        if skip_empty_dst and tile_coord not in dst_tiles:
            continue

        with dst.tile_request(*tile_coord, readonly=False) as dst_tile:

            # Only at this point might the bounding box need to be updated
            dst_changed_bbox = update_bbox(dst_changed_bbox, *tile_coord)

            # Under certain conditions, direct copies and dict manipulation
            # can be used instead of compositing operations.
            cut_off = trim_result and tiles_bbox.crossing(tile_coord)
            full_inner = src_tile is _FULL_TILE and not cut_off
            if full_inner:
                if mode == myplib.CombineNormal and opacity == 1.0:
                    myplib.tile_copy_rgba16_into_rgba16(full_rgba, dst_tile)
                    continue
                elif mode == myplib.CombineDestinationOut and opacity == 1.0:
                    dst_tiles.pop(tile_coord)
                    continue
                elif mode == myplib.CombineDestinationIn and opacity == 1.0:
                    continue
                # Even if opacity != 1.0, we can reuse the full rgba tile
                src_tile_rgba = full_rgba
            else:
                if trim_result:
                    tile_bounds = tiles_bbox.tile_bounds(tile_coord)
                else:
                    tile_bounds = (0, 0, N-1, N-1)
                src_tile_rgba = myplib.rgba_tile_from_alpha_tile(
                    src_tile, *(fill_col + tile_bounds)
                )

            # If alpha locking is enabled in combination with a mode other than
            # CombineNormal, we need to copy the dst tile to mask the result
            if lock_alpha and mode != myplib.CombineSourceAtop:
                mask = np.copy(dst_tile)
                mask_mode = myplib.CombineDestinationAtop
                tile_combine(mode, src_tile_rgba, dst_tile, True, opacity)
                tile_combine(mask_mode, mask, dst_tile, True, 1.0)
            else:
                tile_combine(mode, src_tile_rgba, dst_tile, True, opacity)

    # Handle dst-out and dst-atop: clear untouched tiles
    if mode in [myplib.CombineDestinationIn, myplib.CombineDestinationAtop]:
        for tile_coord in list(dst_tiles.keys()):
            if not handler.run:
                break
            if tile_coord not in filled:
                dst_changed_bbox = update_bbox(
                    dst_changed_bbox, *tile_coord
                )
                with dst.tile_request(*tile_coord, readonly=False):
                    dst_tiles.pop(tile_coord)

    if dst_changed_bbox and handler.run:
        min_tx, min_ty, max_tx, max_ty = dst_changed_bbox
        bbox = (
            min_tx * N, min_ty * N,
            (1 + max_tx - min_tx) * N,
            (1 + max_ty - min_ty) * N,
        )
        # Even for large fills on slow machines, this stage
        # will almost always be too short to even notice.
        # It is not cancellable once entered.
        handler.set_stage(FillHandler.FINISHING)

        # The observers may directly or indirectly use the
        # Gtk API, so the call is scheduled on the gui thread.
        GLib.idle_add(dst.notify_observers, *bbox)


def scanline_fill(handler, src, seed_lists, tiles_bbox, filler):
    """ Perform a scanline fill and return the filled tiles

    Perform a scanline fill using the given starting point and tile,
    with reference to the src surface and given bounding box, using the
    provided filler instance.

    :param handler: updates fill status and permits cancelling
    :type handler: FillHandler
    :param src: Source surface-like object
    :param seed_lists: dictionary, pairing tile coords with lists of seeds
    :type seed_lists: dict
    :param tiles_bbox: Bounding box for the fill
    :type tiles_bbox: lib.fill_common.TileBoundingBox
    :param filler: filler instance performing the per-tile fill operation
    :type filler: myplib.Filler
    :returns: a dictionary of coord->tile mappings for the filled tiles
    """
    # Dict of coord->tile data populated during the fill
    filled = {}

    inv_edges = (
        EDGE.south,
        EDGE.west,
        EDGE.north,
        EDGE.east
    )

    # Starting coordinates + direction of origin (from within)
    tileq = []
    for seed_tile_coord, seeds in iteritems(seed_lists):
        tileq.append((seed_tile_coord, seeds, myplib.edges.none))

    tfs = _TileFillSkipper(tiles_bbox, filler, set({}))

    while len(tileq) > 0 and handler.run:
        tile_coord, seeds, from_dir = tileq.pop(0)
        # Skip if the tile has been fully processed already
        if tile_coord in tfs.final:
            continue
        # Flood-fill one tile
        with src.tile_request(*tile_coord, readonly=True) as src_tile:
            # See if the tile can be skipped
            overflows = tfs.check(tile_coord, src_tile, filled, from_dir)
            if overflows is None:
                if tile_coord not in filled:
                    handler.inc_processed()
                    filled[tile_coord] = np.zeros((N, N), 'uint16')
                overflows = filler.fill(
                    src_tile, filled[tile_coord], seeds,
                    from_dir, *tiles_bbox.tile_bounds(tile_coord)
                )
            else:
                handler.inc_processed()
        enqueue_overflows(tileq, tile_coord, overflows, tiles_bbox, inv_edges)
    return filled


class _TileFillSkipper:
    """Provides checking for, and handling of, uniform tiles"""

    FULL_OVERFLOWS = [
        ((), [(0, N-1)], [(0, N-1)], [(0, N-1)]),         # from north
        ([(0, N-1)], (), [(0, N-1)], [(0, N-1)]),         # from east
        ([(0, N-1)], [(0, N-1)], (), [(0, N-1)]),         # from south
        ([(0, N-1)], [(0, N-1)], [(0, N-1)], ()),         # from west
        ([(0, N-1)], [(0, N-1)], [(0, N-1)], [(0, N-1)])  # from within
    ]

    def __init__(self, tiles_bbox, filler, final):

        self.uniform_tiles = {}
        self.final = final
        self.tiles_bbox = tiles_bbox
        self.filler = filler

    # Dict of alpha->tile, used for uniform non-opaque tile fills
    # NOTE: these are usually not a result of an intentional fill, but
    # clicking a pixel with color very similar to the intended target pixel
    def uniform_tile(self, alpha):
        """ Return a reference to a uniform alpha tile

        If no uniform tile with the given alpha value exists, one is created
        """
        if alpha not in self.uniform_tiles:
            self.uniform_tiles[alpha] = fc.new_full_tile(alpha)
        return self.uniform_tiles[alpha]

    def check(self, tile_coord, src_tile, filled, from_dir):
        """Check if the tile can be handled without using the fill loop.

        The first time the tile is encountered, check if it is uniform
        and if so, handle it immediately depending on whether it is
        fillable or not.

        If the tile can be handled immediately, returns the overflows
        (new seed ranges), otherwise return None to indicate that the
        fill algorithm needs to be invoked.
        """
        if tile_coord in filled or self.tiles_bbox.crossing(tile_coord):
            return None

        # Returns the alpha of the fill for the tile's color if
        # the tile is uniform, otherwise returns None
        is_empty = src_tile is _EMPTY_RGBA
        alpha = self.filler.tile_uniformity(is_empty, src_tile)

        if alpha is None:
            # No shortcut can be taken, create new tile
            return None
        # Tile is uniform, so there is no need to process
        # it again in the fill loop, either set as
        # a uniformly filled alpha tile or skip it if it
        # cannot be filled at all (unlikely, but not impossible)
        self.final.add(tile_coord)
        if alpha == 0:
            return [(), (), (), ()]
        elif alpha == _OPAQUE:
            filled[tile_coord] = _FULL_TILE
        else:
            filled[tile_coord] = self.uniform_tile(alpha)
        return self.FULL_OVERFLOWS[from_dir]


def gap_closing_fill(
        handler, src, seed_lists, tiles_bbox, filler, gap_closing_options):
    """ Fill loop that finds and uses gap data to avoid unwanted leaks

    Gaps are defined as distances of fillable pixels enclosed on two sides
    by unfillable pixels. Each tile considered, and their neighbours, are
    flooded with alpha values based on the target color and threshold values.
    The resulting alphas are then searched for gaps, and the size of these gaps
    are marked in separate tiles - one for each tile filled.
    """

    unseep_queue = []
    filled = {}
    final = set({})

    seed_queue = []
    for seed_tile_coord, seeds in iteritems(seed_lists):
        seed_queue.append((seed_tile_coord, seeds))

    options = gap_closing_options
    max_gap_size = lib.helpers.clamp(options.max_gap_size, 1, TILE_SIZE)
    gc_filler = myplib.GapClosingFiller(max_gap_size, options.retract_seeps)
    gc_handler = _GCTileHandler(final, max_gap_size, tiles_bbox, filler, src)
    total_px = 0
    skip_unseeping = False

    while len(seed_queue) > 0 and handler.run:
        tile_coord, seeds = seed_queue.pop(0)
        if tile_coord in final:
            continue
        # Create distance-data and alpha output tiles for the fill
        # and check if the tile can be skipped directly
        alpha_t, dist_t, overflows = gc_handler.get_gc_data(tile_coord, seeds)
        if overflows:
            handler.inc_processed()
            filled[tile_coord] = _FULL_TILE
        else:
            # Complement data for initial seeds (if they are initial seeds)
            seeds, any_not_max = complement_gc_seeds(seeds, dist_t)
            # If the fill is starting at a point with a detected distance,
            # disable seep retraction - otherwise it is very likely
            # that the result will be completely empty.
            if any_not_max:
                skip_unseeping = True
            # Pixel limits within tiles can vary at the bounding box edges
            px_bounds = tiles_bbox.tile_bounds(tile_coord)
            # Create new output tile if not already present
            if tile_coord not in filled:
                handler.inc_processed()
                filled[tile_coord] = np.zeros((N, N), 'uint16')
            # Run the gap-closing fill for the tile
            result = gc_filler.fill(
                alpha_t, dist_t, filled[tile_coord], seeds, *px_bounds
            )
            overflows = result[0:4]
            fill_edges, px_f = result[4:6]
            # The entire tile was filled, despite potential gaps;
            # replace data w. constant and mark tile as final.
            if px_f == N*N:
                final.add(tile_coord)
            # When seep inversion is enabled, track total pixels filled
            # and coordinates where the fill stopped due to distance conditions
            total_px += px_f
            if not skip_unseeping and fill_edges:
                unseep_queue.append((tile_coord, fill_edges, True))
        # Enqueue overflows, whether skipping or not
        enqueue_overflows(seed_queue, tile_coord, overflows, tiles_bbox)

    # If enabled, pull the fill back into the gaps to stop before them
    if not skip_unseeping and handler.run:
        unseep(
            unseep_queue, filled, gc_filler,
            total_px, tiles_bbox, gc_handler.distances
        )
    return filled


class _GCTileHandler(object):
    """Gap-closing-fill Tile Handler

    Manages input alpha tiles and distance tiles necessary to perform
    gap closing fill operations.
    """
    OVERFLOWS = [
        [(), (EDGE.west,), (EDGE.north,), (EDGE.east,)],
        [(EDGE.south,), (), (EDGE.north,), (EDGE.east,)],
        [(EDGE.south,), (EDGE.west,), (), (EDGE.east,)],
        [(EDGE.south,), (EDGE.west,), (EDGE.north,), ()],
        [(EDGE.south,), (EDGE.west,), (EDGE.north,), (EDGE.east,)],
    ]

    def __init__(self, final, max_gap_size, tiles_bbox, filler, src):
        self._src = src
        self.final = final
        self.distances = dict()
        self._alpha_tiles = dict()
        self._dist_data = None
        self._bbox = tiles_bbox
        self._filler = filler
        self._distbucket = myplib.DistanceBucket(max_gap_size)

    def get_gc_data(self, tile_coord, seeds):
        """Get the data necessary to run a gap-closing fill

        For the given tile coordinate, prepare the data necessary to
        run a gap-closing fill operation for that tile, namely the
        corresponding input alpha tile and distance tile.

        The first time a coordinate is reached, also check if it can be
        skipped directly, and return the overflows if that is the case.

        :returns: (alpha_tile, distance_tile, overflows)
        :rtype: tuple
        """
        if tile_coord not in self.distances:
            # Ensure that alpha data exists for the tile and its neighbours
            grid, all_full = self.alpha_grid(tile_coord)
            # The search is skipped if we have a 9-grid of only full tiles
            # since there cannot be any gaps in that case. Otherwise, if no
            # gaps were found during the search, use a constant distance tile
            if all_full or not self.find_gaps(*grid):
                self.distances[tile_coord] = _GAPLESS_TILE
                # Check if fill can be skipped directly
                can_skip_fill = (
                    (all_full or grid[0] is _FULL_TILE) and
                    not self._bbox.crossing(tile_coord) and
                    gc_seeds_skippable(seeds)
                )
                if can_skip_fill:
                    self.final.add(tile_coord)
                    if isinstance(seeds, list):
                        overflows = self.OVERFLOWS[EDGE.none]
                    else:
                        overflows = self.OVERFLOWS[seeds[0]]
                    return _FULL_TILE, _GAPLESS_TILE, overflows
            else:
                self.distances[tile_coord] = self._dist_data
                self._dist_data = None
        # The distance data is already present, meaning the skip checks have
        # already been tried, no skipping possible.
        return self._alpha_tiles[tile_coord], self.distances[tile_coord], ()

    def find_gaps(self, *grid):
        """Search for and mark gaps, given a nine-grid of alpha tiles

        :param grid: nine-grid of alpha tiles
        :return: True if any gaps were found, otherwise false
        :rtype: bool
        """
        if self._dist_data is None:
            self._dist_data = fc.new_full_tile(INF_DIST)
        return myplib.find_gaps(self._distbucket, self._dist_data, *grid)

    def alpha_grid(self, tile_coord):
        """When needed, create and calculate alpha tiles for distance searching.

        For the tile of the given coordinate, ensure that a corresponding tile
        of alpha values (based on the tolerance function) exists in the
        full_alphas dict for both the tile and all of its neighbors

        :returns: Tuple with the grid and a boolean value indicating
        whether every tile in the grid is the constant full alpha tile
        """
        all_full = True
        alpha_tiles = self._alpha_tiles
        grid = []
        for ntc in fc.nine_grid(tile_coord):
            if ntc not in alpha_tiles:
                with self._src.tile_request(
                        ntc[0], ntc[1], readonly=True
                ) as src_tile:
                    is_empty = src_tile is _EMPTY_RGBA
                    alpha = self._filler.tile_uniformity(is_empty, src_tile)
                    if alpha == _OPAQUE:
                        alpha_tiles[ntc] = _FULL_TILE
                    elif alpha == 0:
                        alpha_tiles[ntc] = _EMPTY_TILE
                    elif alpha:
                        alpha_tiles[ntc] = fc.new_full_tile(alpha)
                    else:
                        alpha_tile = np.empty((N, N), 'uint16')
                        self._filler.flood(src_tile, alpha_tile)
                        alpha_tiles[ntc] = alpha_tile
            tile = alpha_tiles[ntc]
            grid.append(tile)
            all_full = all_full and tile is _FULL_TILE
        return grid, all_full


def unseep(seed_queue, filled, gc_filler, total_px, tiles_bbox, distances):
    """Seep inversion is basically a four-way 0-alpha fill
    with different conditions. It only backs off into the original
    fill and therefore does not require creation of new tiles or use
    of an input alpha tile.
    """
    backup = {}
    while len(seed_queue) > 0:
        tile_coord, seeds, is_initial = seed_queue.pop(0)
        if tile_coord not in distances or tile_coord not in filled:
            continue
        if tile_coord not in backup:
            if filled[tile_coord] is _FULL_TILE:
                backup[tile_coord] = _FULL_TILE
                filled[tile_coord] = fc.new_full_tile(1 << 15)
            else:
                backup[tile_coord] = np.copy(filled[tile_coord])
        result = gc_filler.unseep(
            distances[tile_coord], filled[tile_coord], seeds, is_initial
        )
        overflows = result[0:4]
        num_erased_pixels = result[4]
        total_px -= num_erased_pixels
        enqueue_overflows(
            seed_queue, tile_coord, overflows, tiles_bbox, (False,) * 4
        )
    if total_px <= 0:
        # For small areas, when starting on a distance-marked pixel,
        # backing off may remove the entire fill, in which case we
        # roll back the tiles that were processed
        for tile_coord, tile in iteritems(backup):
            filled[tile_coord] = tile


def complement_gc_seeds(seeds, distance_tile):
    """Add distances to initial seeds, check if all seeds lie on detected gaps

    If the input seeds are not initial seeds, they are returned unchanged.

    Returns a tuple with complemented seeds and a boolean indicating whether
    all seeds lie on detected gaps (this check is only done for initial seeds)

    """
    if isinstance(seeds, list) and len(seeds[0]) < 3:
        # Fetch distance for initial seed coord
        complemented_seeds = []
        any_not_max = False
        for (px, py) in seeds:
            distance = distance_tile[py][px]
            if distance < INF_DIST:
                any_not_max = True
            complemented_seeds.append((px, py, distance))
        return complemented_seeds, any_not_max
    else:
        return seeds, False


def gc_seeds_skippable(seeds):
    return (
        isinstance(seeds, tuple) or  # edge constant - a full edge of seeds
        len(seeds[0]) == 2 or  # initial seeds
        any([s[2] == INF_DIST for s in seeds])  # one seed can fill everything
    )



# This file is part of MyPaint.
# Copyright (C) 2012-2019 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


"""Viewer and editor widgets for palettes."""

# Editor ideas
#   - "Insert lighter/darker copy of row".
#   - repack palette (remove duplicates and blanks)
#   - sort palette by approx. hue+chroma binning, then luma variations


## Imports
from __future__ import division, print_function

import math
import os
import re
import logging
from io import open

from lib.gibindings import Gdk
from lib.gibindings import Gtk
from lib.gibindings import GLib
import cairo
from lib.gettext import C_

from .util import clamp
from lib.palette import Palette
from lib.color import RGBColor
from lib.color import HCYColor
from lib.color import HSVColor
import gui.uicolor
from .adjbases import ColorAdjuster
from .adjbases import ColorAdjusterWidget
from .adjbases import ColorManager
from .adjbases import DATAPATH_PALETTES_SUBDIR
from .combined import CombinedAdjusterPage

from lib.pycompat import unicode


logger = logging.getLogger(__name__)


## Class defs

class PalettePage (CombinedAdjusterPage):
    """User-editable palette, as a `CombinedAdjuster` element."""

    def __init__(self):
        view = PaletteView()
        view.grid.show_matched_color = True
        view.can_select_empty = False
        self._adj = view
        self._edit_dialog = None

    @classmethod
    def get_properties_description(cls):
        return C_(
            "palette panel: properties button tooltip",
            "Palette properties",
        )

    @classmethod
    def get_page_icon_name(cls):
        return "mypaint-tool-color-palette"

    @classmethod
    def get_page_title(cls):
        return C_("palette panel tab tooltip title", "Palette")

    @classmethod
    def get_page_description(cls):
        return C_(
            "palette panel tab tooltip description",
            "Set the color from a loadable, editable palette.",
        )

    def get_page_widget(self):
        """Page widget: returns the PaletteView adjuster widget itself."""
        # FIXME: The PaletteNext and PalettePrev actions of the main
        #        app require access to the PaletteView itself.
        return self._adj

    def set_color_manager(self, manager):
        CombinedAdjusterPage.set_color_manager(self, manager)
        self._adj.set_color_manager(manager)

    def show_properties(self):
        if self._edit_dialog is None:
            toplevel = self._adj.get_toplevel()
            dialog = PaletteEditorDialog(toplevel, self.get_color_manager())
            self._edit_dialog = dialog
        self._edit_dialog.run()


class PaletteEditorDialog (Gtk.Dialog):
    """Dialog for editing, loading and saving the current palette."""

    _UNTITLED_PALETTE_NAME = C_(
        "palette editor dialog: palette name entry",
        "Untitled Palette",
    )

    def __init__(self, parent, target_color_manager):
        Gtk.Dialog.__init__(
            self,
            title=C_("palette editor dialog: title", "Palette Editor"),
            transient_for=parent,
            modal=True,
            destroy_with_parent=True,
        )
        self.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.REJECT)
        self.add_button(Gtk.STOCK_OK, Gtk.ResponseType.ACCEPT)
        self.set_position(Gtk.WindowPosition.MOUSE)

        assert isinstance(target_color_manager, ColorManager)
        #: ColorManager containing the palette to be edited.
        self._target_color_manager = target_color_manager

        view = PaletteView()
        view.set_size_request(400, 300)
        view.grid.show_matched_color = True
        view.grid.can_select_empty = True
        self._view = view

        #: The working ColorManager, holding a working copy of the palette
        #: to be edited.
        self._mgr = ColorManager(prefs={},
                                 datapath=target_color_manager.get_data_path())
        self._mgr.set_color(RGBColor(1, 1, 1))
        view.set_color_manager(self._mgr)

        # Action buttons, positiopned down the right hand side
        action_bbox = Gtk.VButtonBox()
        load_btn = self._load_button = Gtk.Button(stock=Gtk.STOCK_OPEN)
        save_btn = self._save_button = Gtk.Button(stock=Gtk.STOCK_SAVE)
        add_btn = self._add_button = Gtk.Button(stock=Gtk.STOCK_ADD)
        remove_btn = self._remove_button = Gtk.Button(stock=Gtk.STOCK_REMOVE)
        clear_btn = self._clear_button = Gtk.Button(stock=Gtk.STOCK_CLEAR)
        action_bbox.pack_start(load_btn, True, True, 0)
        action_bbox.pack_start(save_btn, True, True, 0)
        action_bbox.pack_start(add_btn, True, True, 0)
        action_bbox.pack_start(remove_btn, True, True, 0)
        action_bbox.pack_start(clear_btn, True, True, 0)
        action_bbox.set_layout(Gtk.ButtonBoxStyle.START)
        load_btn.connect("clicked", self._load_btn_clicked)
        save_btn.connect("clicked", self._save_btn_clicked)
        remove_btn.connect("clicked", self._remove_btn_clicked)
        add_btn.connect("clicked", self._add_btn_clicked)
        clear_btn.connect("clicked", self._clear_btn_clicked)
        load_btn.set_tooltip_text(C_(
            "palette editor dialog: action buttons: tooltips",
            "Load from a GIMP palette file",
        ))
        save_btn.set_tooltip_text(C_(
            "palette editor dialog: action buttons: tooltips",
            "Save to a GIMP palette file",
        ))
        add_btn.set_tooltip_text(C_(
            "palette editor dialog: action buttons: tooltips",
            "Add a new empty swatch",
        ))
        remove_btn.set_tooltip_text(C_(
            "palette editor dialog: action buttons: tooltips",
            "Remove the current swatch",
        ))
        clear_btn.set_tooltip_text(C_(
            "palette editor dialog: action buttons: tooltips",
            "Remove all swatches",
        ))

        # Button initial state and subsequent updates
        remove_btn.set_sensitive(False)
        self._mgr.palette.match_changed += self._palette_match_changed_cb
        self._mgr.palette.info_changed += self._palette_changed_cb
        self._mgr.palette.sequence_changed += self._palette_changed_cb
        self._mgr.palette.color_changed += self._palette_changed_cb

        # Palette name and number of entries
        palette_details_hbox = Gtk.HBox()
        palette_name_label = Gtk.Label(label=C_(
            "palette editor dialog: palette name/title entry: label",
            "Title:",
        ))
        palette_name_label.set_tooltip_text(C_(
            "palette editor dialog: palette name/title entry: tooltip",
            "Name or description for this palette",
        ))
        palette_name_entry = Gtk.Entry()
        palette_name_entry.connect("changed", self._palette_name_changed_cb)
        self._palette_name_entry = palette_name_entry
        self._columns_adj = Gtk.Adjustment(
            value=0, lower=0, upper=99,
            step_increment=1, page_increment=1, page_size=0
        )
        self._columns_adj.connect("value-changed", self._columns_changed_cb)
        columns_label = Gtk.Label(label=C_(
            "palette editor dialog: number-of-columns spinbutton: title",
            "Columns:"
        ))
        columns_label.set_tooltip_text(C_(
            "palette editor dialog: number-of-columns spinbutton: tooltip",
            "Number of columns",
        ))
        columns_spinbutton = Gtk.SpinButton(
            adjustment=self._columns_adj,
            climb_rate=1.5,
            digits=0
        )
        palette_details_hbox.set_spacing(0)
        palette_details_hbox.set_border_width(0)
        palette_details_hbox.pack_start(palette_name_label, False, False, 0)
        palette_details_hbox.pack_start(palette_name_entry, True, True, 6)
        palette_details_hbox.pack_start(columns_label, False, False, 6)
        palette_details_hbox.pack_start(columns_spinbutton, False, False, 0)

        color_name_hbox = Gtk.HBox()
        color_name_label = Gtk.Label(label=C_(
            "palette editor dialog: color name entry: label",
            "Color name:",
        ))
        color_name_label.set_tooltip_text(C_(
            "palette editor dialog: color name entry: tooltip",
            "Current color's name",
        ))
        color_name_entry = Gtk.Entry()
        color_name_entry.connect("changed", self._color_name_changed_cb)
        color_name_entry.set_sensitive(False)
        self._color_name_entry = color_name_entry
        color_name_hbox.set_spacing(6)
        color_name_hbox.pack_start(color_name_label, False, False, 0)
        color_name_hbox.pack_start(color_name_entry, True, True, 0)

        palette_vbox = Gtk.VBox()
        palette_vbox.set_spacing(12)
        palette_vbox.pack_start(palette_details_hbox, False, False, 0)
        palette_vbox.pack_start(view, True, True, 0)
        palette_vbox.pack_start(color_name_hbox, False, False, 0)

        # Dialog contents
        # Main edit area to the left, buttons to the right
        hbox = Gtk.HBox()
        hbox.set_spacing(12)
        hbox.pack_start(palette_vbox, True, True, 0)
        hbox.pack_start(action_bbox, False, False, 0)
        hbox.set_border_width(12)
        self.vbox.pack_start(hbox, True, True, 0)

        # Dialog vbox contents must be shown separately
        for w in self.vbox:
            w.show_all()

        self.connect("response", self._response_cb)
        self.connect("show", self._show_cb)

    def _show_cb(self, widget, *a):
        # Each time the dialog is shown, update with the target
        # palette, for editing.
        self.vbox.show_all()
        palette = self._target_color_manager.palette
        name = palette.get_name()
        if not name:
            name = self._ensure_valid_palette_name()
        self._palette_name_entry.set_text(name)
        self._columns_adj.set_value(palette.get_columns())
        self._mgr.palette.update(palette)

    def _palette_name_changed_cb(self, editable):
        name = editable.get_chars(0, -1)
        if not name:
            name = ""  # note: not None (it'll be stringified)
        pal = self._mgr.palette
        pal.name = unicode(name)

    def _columns_changed_cb(self, adj):
        ncolumns = int(adj.get_value())
        pal = self._mgr.palette
        pal.set_columns(ncolumns)

    def _color_name_changed_cb(self, editable):
        name = editable.get_chars(0, -1)
        palette = self._mgr.palette
        i = palette.match_position
        if i is None:
            return
        old_name = palette.get_color_name(i)
        if not name:
            name = None
        if name != old_name:
            palette.set_color_name(i, name)

    def _response_cb(self, widget, response_id):
        if response_id == Gtk.ResponseType.ACCEPT:
            palette = self._mgr.palette
            target_palette = self._target_color_manager.palette
            target_palette.update(palette)
        self.hide()
        return True

    def _palette_match_changed_cb(self, palette):
        col_name_entry = self._color_name_entry
        i = palette.match_position
        if i is not None:
            col = palette[i]
            if col is not None:
                name = palette.get_color_name(i)
                if name is None:
                    name = ""
                col_name_entry.set_sensitive(True)
                col_name_entry.set_text(name)
            else:
                col_name_entry.set_sensitive(False)
                col_name_entry.set_text(C_(
                    "palette editor dialog: color name entry",
                    "Empty palette slot",
                ))
        else:
            col_name_entry.set_sensitive(False)
            col_name_entry.set_text("")
        self._update_buttons()

    def _update_buttons(self):
        palette = self._mgr.palette
        emptyish = len(palette) == 0
        if len(palette) == 1:
            if palette[0] is None:
                emptyish = True
        can_save = not emptyish
        can_clear = not emptyish
        can_remove = True
        if emptyish or self._mgr.palette.match_position is None:
            can_remove = False
        self._save_button.set_sensitive(can_save)
        self._remove_button.set_sensitive(can_remove)
        self._clear_button.set_sensitive(can_clear)

    def _palette_changed_cb(self, palette, *args, **kwargs):
        new_name = palette.get_name()
        if new_name is None:
            new_name = self._ensure_valid_palette_name()
        old_name = self._palette_name_entry.get_chars(0, -1)
        if old_name != new_name:
            self._palette_name_entry.set_text(new_name)
        self._columns_adj.set_value(palette.get_columns())
        self._update_buttons()

    def _add_btn_clicked(self, button):
        palette = self._mgr.palette
        i = palette.match_position
        if i is None:
            i = len(palette)
            palette.append(None)
            palette.match_position = i
        else:
            palette.insert(i, None)

    def _remove_btn_clicked(self, button):
        palette = self._mgr.palette
        i = palette.match_position
        if i >= 0 and i < len(palette):
            palette.pop(i)
            if len(palette) == 0:
                palette.append(None)

    def _load_btn_clicked(self, button):
        preview = _PalettePreview()
        manager = self._target_color_manager
        datapath = manager.get_data_path()
        palettes_dir = os.path.join(datapath, DATAPATH_PALETTES_SUBDIR)
        palette = palette_load_via_dialog(
            title=C_("palette load dialog: title", "Load palette"),
            parent=self,
            preview=preview,
            shortcuts=[palettes_dir],
        )
        if palette is not None:
            self._mgr.palette.update(palette)
            self._ensure_valid_palette_name()

    def _save_btn_clicked(self, button):
        preview = _PalettePreview()
        palette_save_via_dialog(
            self._mgr.palette,
            title=C_("palette save dialog: title", "Save palette"),
            parent=self,
            preview=preview,
        )

    def _clear_btn_clicked(self, button):
        pal = self._mgr.palette
        pal.clear()
        self._ensure_valid_palette_name()

    def _ensure_valid_palette_name(self):
        pal = self._mgr.palette
        if not pal.name:
            pal.name = self._UNTITLED_PALETTE_NAME
        return pal.name


class PaletteView (ColorAdjuster, Gtk.ScrolledWindow):
    """Scrollable view of a palette.

    Palette entries can be clicked to select the color, and all instances of
    the current shared color in the palette are highlighted.

    """

    ## Sizing constraint constants
    _MIN_HEIGHT = 32
    _MIN_WIDTH = 150
    _MAX_NATURAL_HEIGHT = 300
    _MAX_NATURAL_WIDTH = 300

    def __init__(self):
        Gtk.ScrolledWindow.__init__(self)
        self.grid = _PaletteGridLayout()
        self.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        self.add(self.grid)

    def set_color_manager(self, mgr):
        self.grid.set_color_manager(mgr)
        ColorAdjuster.set_color_manager(self, mgr)

    ## Sizing boilerplate
    # Reflect what the embedded grid widget tells us, but limit its natural
    # size to something sensible. Huge palettes make huge grids...

    def do_get_request_mode(self):
        return self.grid.get_request_mode()

    def do_get_preferred_width(self):
        gminw, gnatw = self.grid.get_preferred_width()
        minw = self._MIN_WIDTH
        natw = min(gnatw, self._MAX_NATURAL_WIDTH)
        return minw, max(minw, natw)

    def do_get_preferred_height(self):
        gminh, gnath = self.grid.get_preferred_height()
        minh = self._MIN_HEIGHT
        nath = min(gnath, self._MAX_NATURAL_HEIGHT)
        return minh, max(minh, nath)

    def do_get_preferred_width_for_height(self, height):
        gminw, gnatw = self.grid.get_preferred_width_for_height(height)
        minw = self._MIN_WIDTH
        natw = min(gnatw, self._MAX_NATURAL_WIDTH)
        return minw, max(minw, natw)

    def do_get_preferred_height_for_width(self, width):
        gminh, gnath = self.grid.get_preferred_height_for_width(width)
        minh = self._MIN_HEIGHT
        nath = min(gnath, self._MAX_NATURAL_HEIGHT)
        return minh, max(minh, nath)


class _PalettePreview (Gtk.DrawingArea):
    """Preview-only palette view."""

    _palette = None

    def __init__(self):
        Gtk.DrawingArea.__init__(self)
        self.connect("draw", self._draw_cb)
        self.set_size_request(128, 256)

    def _draw_cb(self, widget, cr):
        if self._palette is None:
            return
        alloc = widget.get_allocation()
        w, h = alloc.width, alloc.height
        s_max = 16  # min(w, h)
        s_min = 4
        ncolumns = self._palette.get_columns()
        ncolors = len(self._palette)
        if ncolors == 0:
            return
        if not ncolumns == 0:
            s = w / ncolumns
            s = clamp(s, s_min, s_max)
            s = int(s)
            if (s * ncolumns) > w:
                ncolumns = 0
        if ncolumns == 0:
            s = math.sqrt((w * h) / ncolors)
            s = clamp(s, s_min, s_max)
            s = int(s)
            ncolumns = max(1, int(w // s))
        nrows = int(ncolors // ncolumns)
        if ncolors % ncolumns != 0:
            nrows += 1
        nrows = max(1, nrows)
        dx, dy = 0, 0
        if (nrows * s) < h:
            dy = int(h - (nrows * s)) // 2
        if (ncolumns * s) < w:
            dx = int(w - (ncolumns * s)) // 2
        bg_color = _widget_get_bg_color(self)
        _palette_render(self._palette, cr, rows=nrows, columns=ncolumns,
                        swatch_size=s, bg_color=bg_color,
                        offset_x=dx, offset_y=dy,
                        rtl=False)

    def set_palette(self, palette):
        self._palette = palette
        self.queue_draw()


def _widget_get_bg_color(widget):
    """Valid background color from the first ancestor widget having one

    Workaround for some widget arrangements in Adwaita for 3.14 having
    null background colors. Fallback is a medium grey, which should be
    acceptable with most styles.

    """
    while widget is not None:
        state = widget.get_state_flags()
        style_context = widget.get_style_context()
        bg_rgba = style_context.get_background_color(state)
        if bg_rgba.alpha != 0:
            return gui.uicolor.from_gdk_rgba(bg_rgba)
        widget = widget.get_parent()
    return RGBColor(0.5, 0.5, 0.5)


class _PaletteGridLayout (ColorAdjusterWidget):
    """The palette layout embedded in a scrolling PaletteView.
    """

    ## Class settings
    IS_DRAG_SOURCE = True
    HAS_DETAILS_DIALOG = True
    STATIC_TOOLTIP_TEXT = C_(
        "palette view",
        "Color swatch palette.\nDrop colors here,\ndrag them to organize.",
    )
    ALLOW_HCY_TWEAKING = False   # Interacts badly with menus

    ## Layout constants
    _SWATCH_SIZE_MIN = 8
    _SWATCH_SIZE_MAX = 50
    _SWATCH_SIZE_NOMINAL = 20
    _PREFERRED_COLUMNS = 5  #: Preferred width in cells for free-flow mode.

    def __init__(self):
        ColorAdjusterWidget.__init__(self)
        # Sizing
        s = self._SWATCH_SIZE_NOMINAL
        self.set_size_request(s, s)
        self.connect("size-allocate", self._size_alloc_cb)
        #: Highlight the currently matched color
        self.show_matched_color = False
        #: User can click on empty slots
        self.can_select_empty = False
        # Current index
        self.connect("button-press-event", self._button_press_cb)
        self.connect_after("button-release-event", self._button_release_cb)
        # Dragging
        self.connect("motion-notify-event", self._motion_notify_cb)
        self.add_events(Gdk.EventMask.POINTER_MOTION_MASK)
        # Tooltips
        self._tooltip_index = None
        self.set_has_tooltip(True)
        # Target markers while dragging or invoking the context menu
        self._insert_target_index = None
        # Cached layout details
        self._rows = None
        self._columns = None
        self._last_palette_columns = None
        self._swatch_size = self._SWATCH_SIZE_NOMINAL

    def _size_alloc_cb(self, widget, alloc):
        """Caches layout details after size negotiation.
        """
        width = alloc.width
        height = alloc.height
        ncolors, nrows, ncolumns = self._get_palette_dimensions()
        if nrows and ncolumns:
            # Fitted to the major dimension
            size = int(min(width / ncolumns, height / nrows))
            size = self._constrain_swatch_size(size)
        else:
            # Free-flowing
            if ncolors > 0:
                size = int(math.sqrt((width * height) / ncolors))
                size = self._constrain_swatch_size(size)
                ncolumns = max(1, min(ncolors, width // size))
                nrows = max(1, int(ncolors // ncolumns))
                if int(ncolors % ncolumns) > 0:
                    nrows += 1
                if nrows * size > height or ncolumns * size > width:
                    size = max(1, min(int(height // nrows),
                                      int(width // ncolumns)))
                    size = self._constrain_swatch_size(size)
                    ncolumns = max(1, min(ncolors, width // size))
                    nrows = max(1, int(ncolors // ncolumns))
                    if int(ncolors % ncolumns) > 0:
                        nrows += 1
            else:
                nrows = 0
                ncolumns = 0
                size = self._SWATCH_SIZE_NOMINAL
        self._rows = nrows
        self._columns = ncolumns
        self._swatch_size = size

    ## Palette monitoring

    def set_color_manager(self, mgr):
        ColorAdjusterWidget.set_color_manager(self, mgr)
        # Could be smarter about these: probably no need to redraw on
        # every little change.
        mgr.palette.info_changed += self._palette_changed_cb
        mgr.palette.match_changed += self._palette_changed_cb
        mgr.palette.sequence_changed += self._palette_changed_cb
        mgr.palette.color_changed += self._palette_changed_cb

    def _palette_changed_cb(self, palette, *args, **kwargs):
        """Called after each change made to the palette."""
        # Determine if the layout has changed since the last time the palette
        # was draw.
        layout_changed = False
        if None in (self._rows, self._columns):
            logger.debug("layout changed: null preexisting layout info")
            layout_changed = True
        else:
            if palette.columns != self._last_palette_columns:
                layout_changed = True
                logger.debug("layout changed: different number of columns")
            else:
                ncells = self._rows * self._columns
                ncolors = len(palette)
                if ncolors > ncells or ncolors <= ncells - self._columns:
                    logger.debug("layout changed: cannot fit palette into "
                                 "currently calculated space")
                    layout_changed = True
        # Queue a resize (and an implicit redraw) if the layout has changed,
        # or just a redraw.
        if layout_changed:
            logger.debug("queuing full resize")
            self._rows = None
            self._columns = None
            self.queue_resize()
            self._insert_target_index = None
            self._tooltip_index = None
        else:
            logger.debug("layout unchanged, queuing redraw")
            self.queue_draw()
        self._last_palette_columns = palette.columns

    ## Pointer event handling

    def _motion_notify_cb(self, widget, event):
        x, y = event.x, event.y
        i = self.get_index_at_pos(x, y)
        # Set the tooltip.
        # Passing the tooltip through a value of None is necessary for its
        # position on the screen to be updated to where the pointer is. Setting
        # it to None, and then to the desired value must happen in two separate
        # events for the tooltip window position update to be honoured.
        if i is None:
            # Not over a color, so use the static default
            if self._tooltip_index not in (-1, -2):
                # First such event: reset the tooltip.
                self._tooltip_index = -1
                self.set_has_tooltip(False)
                self.set_tooltip_text("")
            elif self._tooltip_index != -2:
                # Second event over a non-color: set the tooltip text.
                self._tooltip_index = -2
                self.set_has_tooltip(True)
                self.set_tooltip_text(self.STATIC_TOOLTIP_TEXT)
        elif self._tooltip_index != i:
            # Mouse pointer has moved to a different color, or away
            # from the two states above.
            if self._tooltip_index is not None:
                # First event for this i: reset the tooltip.
                self._tooltip_index = None
                self.set_has_tooltip(False)
                self.set_tooltip_text("")
            else:
                # Second event for this i: set the desired tooltip text.
                self._tooltip_index = i
                mgr = self.get_color_manager()
                tip = mgr.palette.get_color_name(i)
                color = mgr.palette.get_color(i)
                if color is None:
                    tip = C_(
                        "palette view",
                        "Empty palette slot (drag a color here)",
                    )
                elif tip is None or tip.strip() == "":
                    tip = ""  # Anonymous colors don't get tooltips
                self.set_has_tooltip(True)
                self.set_tooltip_text(tip)

    def _button_press_cb(self, widget, event):
        """Handle button presses."""
        # The base class has a separate handler which
        # changes the managed colour, so don't need to do that here.
        if event.type != Gdk.EventType.BUTTON_PRESS:
            return False
        # Move the highlight
        x, y = event.x, event.y
        i = self.get_index_at_pos(x, y, nearest=False)
        mgr = self.get_color_manager()
        is_empty = mgr.palette.get_color(i) is None
        if event.button == 1:
            if not (is_empty and not self.can_select_empty):
                mgr.palette.set_match_position(i)
                mgr.palette.set_match_is_approx(False)
            return False
        # Button 3 shows a menu
        if event.button != 3:
            return False
        self._popup_context_menu(event)

    def _popup_context_menu(self, event):
        x, y = event.x, event.y
        i = self.get_index_at_pos(x, y, nearest=True, insert=True)
        mx, my = self.get_position_for_index(i)
        mx = event.x_root - x + mx + self._swatch_size
        my = event.y_root - y + my + self._swatch_size
        menu = self._get_context_menu(i)
        menu.show_all()
        menu.popup(
            parent_menu_shell = None,
            parent_menu_item = None,
            func = lambda *a: (mx, my, True),
            data = None,
            button = event.button,
            activate_time = event.time,
        )
        self._insert_target_index = i
        self.queue_draw()
        return False

    def _get_empty_range(self, index):
        """Returns the populated start and end of a range of empty slots

        Returns the indices of two populated swatches around the target
        swatch, or None if there's no run of one or more empty slots
        between them.

        """
        start_index = None
        end_index = None
        palette = self.get_color_manager().palette
        if palette[index] is not None:
            return None
        i = index
        while i >= 0:
            i -= 1
            if palette[i] is not None:
                start_index = i
                break
        i = index
        while i < len(palette):
            i += 1
            if palette[i] is not None:
                end_index = i
                break
        if None not in (start_index, end_index):
            assert start_index < end_index
            if start_index < end_index - 1:
                return (start_index, end_index)
        return None

    def _get_context_menu(self, i):
        menu = Gtk.Menu()
        menu.connect_after("deactivate", self._context_menu_deactivate_cb)
        palette = self.get_color_manager().palette
        empty_range = self._get_empty_range(i)
        item_defs = [
            (
                # TRANSLATORS: inserting gaps (empty color swatches)
                C_("palette view: context menu", "Add Empty Slot"),
                self._insert_empty_slot_cb,
                True,
                [i],
            ),
            (
                # TRANSLATORS: inserting gaps (empty color swatches)
                C_("palette view: context menu", "Insert Row"),
                self._insert_empty_row_cb,
                True,
                [i],
            ),
            (
                # TRANSLATORS: inserting gaps (empty color swatches)
                C_("palette view: context menu", "Insert Column"),
                self._insert_empty_column_cb,
                bool(palette.get_columns()),
                [i],
            ),
            None,
            (
                # TRANSLATORS: Color interpolations
                C_("palette view: context menu", "Fill Gap (RGB)"),
                self._interpolate_empty_range_cb,
                bool(empty_range),
                [RGBColor, empty_range],
            ),
            (
                # TRANSLATORS: Color interpolations
                C_("palette view: context menu", "Fill Gap (HCY)"),
                self._interpolate_empty_range_cb,
                bool(empty_range),
                [HCYColor, empty_range],
            ),
            (
                # TRANSLATORS: Color interpolations
                C_("palette view: context menu", "Fill Gap (HSV)"),
                self._interpolate_empty_range_cb,
                bool(empty_range),
                [HSVColor, empty_range],
            ),
        ]
        for item_def in item_defs:
            if not item_def:
                item = Gtk.SeparatorMenuItem()
            else:
                label_str, activate_cb, sensitive, args = item_def
                item = Gtk.MenuItem()
                item.set_label(label_str)
                if activate_cb:
                    item.connect("activate", activate_cb, *args)
                item.set_sensitive(sensitive)
            menu.append(item)
        menu.attach_to_widget(self)
        return menu

    def _button_release_cb(self, widget, event):
        pass

    ## Context menu handlers

    def _context_menu_deactivate_cb(self, menu):
        self._insert_target_index = None
        self.queue_draw()
        GLib.idle_add(menu.destroy)

    def _insert_empty_slot_cb(self, menuitem, target_i):
        palette = self.get_color_manager().palette
        palette.insert(target_i, None)

    def _insert_empty_row_cb(self, menuitem, target_i):
        row_start_i = (target_i // self._columns) * self._columns
        palette = self.get_color_manager().palette
        for i in range(palette.get_columns()):
            palette.insert(row_start_i, None)

    def _insert_empty_column_cb(self, menuitem, target_i):
        palette = self.get_color_manager().palette
        assert palette.get_columns(), \
            "Can't insert columns into a free-flowing palette"
        row_di = target_i % self._columns
        columns_new = palette.get_columns() + 1
        r = 0
        while r * columns_new < len(palette):
            i = r * columns_new + row_di
            if i >= len(palette):
                break
            palette.insert(i, None)
            r += 1
        palette.set_columns(columns_new)

    def _interpolate_empty_range_cb(self, menuitem, color_class, range):
        i0, ix = range
        palette = self.get_color_manager().palette
        c0 = color_class(color=palette[i0])
        cx = color_class(color=palette[ix])
        nsteps = ix - i0 + 1
        if nsteps < 3:
            return
        interpolated = list(c0.interpolate(cx, nsteps))
        assert len(interpolated) == nsteps
        interpolated.pop(0)
        interpolated.pop(-1)
        for i, c in enumerate(interpolated):
            palette[i0 + 1 + i] = c

    ## Dimensions and sizing

    @classmethod
    def _constrain_swatch_size(cls, size):
        size = min(cls._SWATCH_SIZE_MAX, max(cls._SWATCH_SIZE_MIN, size))
        # Restrict to multiples of 2 for patterns, plus one for the border
        if size % 2 == 0:
            size -= 1
        return size

    def _get_palette_dimensions(self):
        """Normalized palette dimensions: (ncolors, nrows, ncolumns).

        Row and columns figures are None if the layout is to be free-flowing.

        """
        mgr = self.get_color_manager()
        ncolumns = mgr.palette.get_columns()
        ncolors = len(mgr.palette)
        if ncolumns is None or ncolumns < 1:
            nrows = None
            ncolumns = None
        else:
            ncolumns = int(ncolumns)
            if ncolors > 0:
                ncolumns = min(ncolumns, ncolors)
                nrows = max(1, int(ncolors // ncolumns))
                if int(ncolors % ncolumns) > 0:
                    nrows += 1
            else:
                ncolumns = 1
                nrows = 1
        return (ncolors, nrows, ncolumns)

    def do_get_request_mode(self):
        """GtkWidget size negotiation implementation
        """
        ncolors, nrows, ncolumns = self._get_palette_dimensions()
        mode = Gtk.SizeRequestMode.HEIGHT_FOR_WIDTH
        if nrows and ncolumns:
            if nrows > ncolumns:
                mode = Gtk.SizeRequestMode.WIDTH_FOR_HEIGHT
        return mode

    def do_get_preferred_width(self):
        """GtkWidget size negotiation implementation.
        """
        ncolors, nrows, ncolumns = self._get_palette_dimensions()
        if ncolumns and ncolumns:
            # Horizontal fit, assume rows <= columns
            min_w = self._SWATCH_SIZE_MIN * ncolumns
            nat_w = self._SWATCH_SIZE_NOMINAL * ncolumns
        else:
            # Free-flowing, across and then down
            ncolumns = max(1, min(self._PREFERRED_COLUMNS, ncolors))
            min_w = self._SWATCH_SIZE_MIN
            nat_w = self._SWATCH_SIZE_NOMINAL * ncolumns
        return min_w, max(min_w, nat_w)

    def do_get_preferred_height_for_width(self, width):
        """GtkWidget size negotiation implementation.
        """
        ncolors, nrows, ncolumns = self._get_palette_dimensions()
        if nrows and ncolumns:
            # Horizontal fit
            swatch_size = self._constrain_swatch_size(int(width // ncolumns))
            min_h = self._SWATCH_SIZE_MIN * nrows
            nat_h = swatch_size * nrows
        else:
            # Free-flowing, across and then down
            # Since s = sqrt((w*h)/n),
            min_h = int(((self._SWATCH_SIZE_MIN ** 2) * ncolors) // width)
            nat_h = int(((self._SWATCH_SIZE_NOMINAL ** 2) * ncolors) // width)
        return min_h, max(min_h, nat_h)

    def do_get_preferred_height(self):
        """GtkWidget size negotiation implementation.
        """
        ncolors, nrows, ncolumns = self._get_palette_dimensions()
        if nrows and ncolumns:
            # Vertical fit, assume rows > columns
            min_h = self._SWATCH_SIZE_MIN * nrows
            nat_h = self._SWATCH_SIZE_NOMINAL * nrows
        else:
            # Height required for our own minimum width (note do_())
            min_w, nat_w = self.do_get_preferred_width()
            min_h, nat_h = self.do_get_preferred_height_for_width(min_w)
        return min_h, max(min_h, nat_h)

    def do_get_preferred_width_for_height(self, height):
        """GtkWidget size negotiation implementation.
        """
        ncolors, nrows, ncolumns = self._get_palette_dimensions()
        if nrows and ncolumns:
            # Vertical fit
            swatch_size = self._constrain_swatch_size(int(height // nrows))
            min_w = self._SWATCH_SIZE_MIN * ncolumns
            nat_w = swatch_size * ncolumns
        else:
            # Just the minimum and natural width (note do_())
            min_w, nat_w = self.do_get_preferred_width()
        return min_w, max(min_w, nat_w)

    def _get_background_size(self):
        # HACK. it's quicker for this widget to render in the foreground
        return 1, 1

    def get_background_validity(self):
        return 1

    def render_background_cb(self, cr, wd, ht):
        return

    def _paint_palette_layout(self, cr):
        mgr = self.get_color_manager()
        if mgr.palette is None:
            return
        bg_col = _widget_get_bg_color(self)
        dx, dy = self.get_painting_offset()
        _palette_render(mgr.palette, cr,
                        rows=self._rows, columns=self._columns,
                        swatch_size=self._swatch_size,
                        bg_color=bg_col,
                        offset_x=dx, offset_y=dy,
                        rtl=False)

    def _paint_marker(self, cr, x, y, insert=False,
                      bg_rgb=(0, 0, 0), fg_rgb=(1, 1, 1),
                      bg_dash=[1, 2], fg_dash=[1, 2],
                      bg_width=2, fg_width=1):
        cr.save()
        cr.set_line_join(cairo.LINE_JOIN_ROUND)
        cr.set_line_cap(cairo.LINE_CAP_ROUND)
        size = self._swatch_size
        w = h = size
        # Background (shadow)
        cr.set_source_rgb(*bg_rgb)
        cr.set_line_width(bg_width)
        cr.rectangle(x, y, w - 1, h - 1)

        cr.set_dash(bg_dash)
        cr.stroke_preserve()

        # Foreground
        cr.set_line_width(fg_width)
        cr.set_dash(fg_dash)
        cr.set_source_rgb(*fg_rgb)
        cr.stroke()
        cr.restore()

    def paint_foreground_cb(self, cr, wd, ht):
        mgr = self.get_color_manager()
        if len(mgr.palette) < 1:
            return

        # Palette cells
        self._paint_palette_layout(cr)

        # Highlights
        cr.set_line_cap(cairo.LINE_CAP_SQUARE)

        # Target marker
        if self._insert_target_index is not None:
            i = self._insert_target_index
            x, y = self.get_position_for_index(i)
            insert = mgr.palette.get_color(i) is not None
            self._paint_marker(cr, x, y, insert=insert)

        # Position of the previous click
        elif self.show_matched_color:
            i = mgr.palette.match_position
            if i is not None:
                x, y = self.get_position_for_index(i)
                marker_args = [cr, x, y]
                marker_kw = dict(bg_width=3, fg_width=1,
                                 bg_dash=[2, 3], fg_dash=[2, 3])
                if not mgr.palette.match_is_approx:
                    marker_kw.update(dict(bg_width=4, fg_width=1))
                self._paint_marker(*marker_args, **marker_kw)

    def get_position_for_index(self, i):
        """Gets the X and Y positions for a color cell at the given index"""
        if None in (self._rows, self._columns):
            return 0, 0
        dx, dy = self.get_painting_offset()
        s_w = s_h = self._swatch_size
        c = i % self._columns
        r = int(i // self._columns)
        x = 0.5 + (c * s_w)
        y = 0.5 + (r * s_h)
        return (x + dx, y + dy)

    def get_painting_offset(self):
        if None in (self._rows, self._columns):
            return 0, 0
        sw = sh = self._swatch_size
        l_wd = sw * self._columns
        l_ht = sh * self._rows
        alloc = self.get_allocation()
        wd, ht = alloc.width, alloc.height
        dx, dy = 0, 0
        if l_wd < wd:
            dx = (wd - l_wd) / 2.0
        if l_ht < ht:
            dy = (ht - l_ht) / 2.0
        return 1 + int(dx), 1 + int(dy)

    def get_color_at_position(self, x, y):
        i = self.get_index_at_pos(x, y)
        if i is not None:
            mgr = self.get_color_manager()
            col = mgr.palette.get_color(i)
            if col is None:
                return None
            return col

    def set_color_at_position(self, x, y, color):
        i = self.get_index_at_pos(x, y)
        mgr = self.get_color_manager()
        if i is None:
            mgr.palette.append(color)
        else:
            mgr.palette[i] = color
        ColorAdjusterWidget.set_color_at_position(self, x, y, color)

    def get_index_at_pos(self, x, y, nearest=False, insert=False):
        """Convert a position to a palette index

        :param int x: X coord, in widget pixels
        :param int y: Y coord, in widget pixels
        :param bool nearest: Pick nearest index if (x, y) lies outside
        :param bool insert: Get an insertion index (requires nearest)
        :rtype: int
        :returns: An index, or None

        The returned index value may be None. Insertion indices are not
        guaranteed to identify existing entries.

        """
        mgr = self.get_color_manager()
        if mgr.palette is None:
            return None
        if None in (self._rows, self._columns):
            return None
        dx, dy = self.get_painting_offset()
        s_wd = s_ht = self._swatch_size
        # Calculate a raw row and column
        r = int((y - dy) // s_ht)
        c = int((x - dx) // s_wd)
        # Check position is within range, or constrain for nearest
        if r < 0:
            if not nearest:
                return None
            r = 0
        elif r >= self._rows:
            if not nearest:
                return None
            r = self._rows - 1
        if c < 0:
            if not nearest:
                return None
            c = 0
        elif c >= self._columns:
            if not nearest:
                return None
            c = self._columns - 1
        # Index range check too: the last row may not be fully populated
        i = (r * self._columns) + c
        if i >= len(mgr.palette):
            if not nearest:
                return None
            i = len(mgr.palette)
            if not insert:
                i -= 1
        return i

    ## Drag handling overrides

    def drag_motion_cb(self, widget, context, x, y, t):
        if "application/x-color" not in map(str, context.list_targets()):
            return False

        # Default action: copy means insert or overwrite
        action = Gdk.DragAction.COPY

        # Update the insertion marker
        i = self.get_index_at_pos(x, y)
        if i != self._insert_target_index:
            self.queue_draw()
        self._insert_target_index = i

        # Dragging around inside the widget implies moving, by default
        source_widget = Gtk.drag_get_source_widget(context)
        if source_widget is self:
            action = Gdk.DragAction.MOVE
            if i is None:
                action = Gdk.DragAction.DEFAULT  # it'll be ignored
            else:
                mgr = self.get_color_manager()
                if mgr.palette.get_color(i) is None:
                    # Empty swatch, convert moves to copies
                    action = Gdk.DragAction.COPY
                    # TODO: record this as a target range for redraws,
                    # and reset _insert_target_index

        # Cursor and status update
        Gdk.drag_status(context, action, t)

    def drag_data_received_cb(self, widget, context, x, y,
                              selection, info, t):
        if "application/x-color" not in map(str, context.list_targets()):
            return False
        data = selection.get_data()
        data_type = selection.get_data_type()
        fmt = selection.get_format()
        logger.debug("drag-data-received: got type=%r", data_type)
        logger.debug("drag-data-received: got fmt=%r", fmt)
        logger.debug("drag-data-received: got data=%r len=%r", data, len(data))
        color = gui.uicolor.from_drag_data(data)
        target_index = self.get_index_at_pos(x, y)

        mgr = self.get_color_manager()
        if Gtk.drag_get_source_widget(context) is self:
            # Move/copy
            current_index = mgr.palette.match_position
            logger.debug("Move/copy %r -> %r", current_index, target_index)
            assert current_index is not None
            mgr.palette.reposition(current_index, target_index)
        else:
            if target_index is None:
                # Append if the drop wasn't over a swatch
                target_index = len(mgr.palette)
            else:
                # Insert before populated swatches, or overwrite empties
                if mgr.palette.get_color(target_index) is None:
                    mgr.palette.pop(target_index)
            mgr.palette.insert(target_index, color)
        self._insert_target_index = None
        self.queue_draw()
        context.finish(True, True, t)
        self.set_managed_color(color)
        mgr.palette.set_match_position(target_index)

    def drag_end_cb(self, widget, context):
        self._insert_target_index = None
        self.queue_draw()

    def drag_leave_cb(self, widget, context, time):
        self._insert_target_index = None
        self.queue_draw()


## Loading and saving of palettes via a dialog


def palette_load_via_dialog(title, parent=None, preview=None,
                            shortcuts=None):
    """Runs a file chooser dialog, returning a palette or `None`.

    The dialog is both modal and blocking. A new `Palette` object is returned
    if the load was successful. The value `None` is returned otherwise.

    :param parent: specifies the parent window
    :param title: dialog title
    :param preview: any preview widget with a ``set_palette()`` method
    :param shortcuts: optional list of shortcut folders

    """
    dialog = Gtk.FileChooserDialog(
        title=title,
        transient_for=parent,
        action=Gtk.FileChooserAction.OPEN,
    )
    dialog.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.REJECT)
    dialog.add_button(Gtk.STOCK_OPEN, Gtk.ResponseType.ACCEPT)
    if preview is not None:
        dialog.set_preview_widget(preview)
        dialog.connect("update-preview",
                       _palette_loadsave_dialog_update_preview_cb,
                       preview)
    if shortcuts is not None:
        for shortcut in shortcuts:
            dialog.add_shortcut_folder(shortcut)
    dialog.set_do_overwrite_confirmation(True)
    filter = Gtk.FileFilter()
    filter.add_pattern("*.gpl")
    filter.set_name(C_(
        "palette load dialog: filters",
        "GIMP palette file (*.gpl)",
    ))
    dialog.add_filter(filter)
    filter = Gtk.FileFilter()
    filter.add_pattern("*")
    filter.set_name(C_(
        "palette load dialog: filters",
        "All files (*)",
    ))
    dialog.add_filter(filter)
    response_id = dialog.run()
    palette = None
    if response_id == Gtk.ResponseType.ACCEPT:
        filename = dialog.get_filename()
        logger.info("Loading palette from %r", filename)
        palette = Palette(filename=filename)
    dialog.destroy()
    return palette


def palette_save_via_dialog(palette, title, parent=None, preview=None):
    """Runs a file chooser dialog for saving.

    The dialog is both modal and blocking. Returns True if the file was saved
    successfully.

    :paraqm palette: the palette to save
    :param parent: specifies the parent window
    :param title: dialog title
    :param preview: any preview widget with a ``set_palette()`` method

    """
    dialog = Gtk.FileChooserDialog(
        title=title,
        transient_for=parent,
        action=Gtk.FileChooserAction.SAVE,
    )
    dialog.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.REJECT)
    dialog.add_button(Gtk.STOCK_SAVE, Gtk.ResponseType.ACCEPT)
    if preview is not None:
        dialog.set_preview_widget(preview)
        dialog.connect("update-preview",
                       _palette_loadsave_dialog_update_preview_cb,
                       preview)
    dialog.set_do_overwrite_confirmation(True)
    filter = Gtk.FileFilter()
    filter.add_pattern("*.gpl")
    filter.set_name(C_(
        "palette save dialog: filters",
        "GIMP palette file (*.gpl)",
    ))
    dialog.add_filter(filter)
    filter = Gtk.FileFilter()
    filter.add_pattern("*")
    filter.set_name(C_(
        "palette save dialog: filters",
        "All files (*)",
    ))
    dialog.add_filter(filter)
    response_id = dialog.run()
    result = False
    if response_id == Gtk.ResponseType.ACCEPT:
        filename = dialog.get_filename()
        filename = re.sub(r'[.]?(?:[Gg][Pp][Ll])?$', "", filename)
        filename += ".gpl"
        logger.info("Saving palette to %r", filename)
        # FIXME: this can overwrite files without prompting the user, if
        # the name hacking above changed the filename.  Should do the name
        # tweak within the dialog somehow and get that to confirm.
        with open(filename, 'w', encoding="utf-8", errors="replace") as fp:
            palette.save(fp)
            fp.flush()
        result = True
    dialog.destroy()
    return result


def _palette_loadsave_dialog_update_preview_cb(dialog, preview):
    """Updates the preview widget when loading/saving palettes via dialog"""
    filename = dialog.get_preview_filename()
    palette = None
    if filename is not None and os.path.isfile(filename):
        try:
            palette = Palette(filename=filename)
        except Exception as ex:
            logger.warning("Couldn't update preview widget: %s", str(ex))
            return
    if palette is not None and len(palette) > 0:
        dialog.set_preview_widget_active(True)
        preview.set_palette(palette)
        preview.queue_draw()
    else:
        dialog.set_preview_widget_active(False)


## Palette rendering using Cairo


def _palette_render(palette, cr, rows, columns, swatch_size,
                    bg_color, offset_x=0, offset_y=0,
                    rtl=False):
    """Renders a Palette according to a precalculated grid.

    :param cr: a Cairo context
    :param int rows: number of rows in the layout
    :param int columns: number of columns in the layout
    :param int swatch_size: size of each swatch, in pixels
    :param lib.color.UIColor bg_color: color used when rendering the patterned
                      placeholder for an empty palette slot.
    :param bool rtl: layout direction: set to True to render right to left,
                 instead of left to right. Currently ignored.
    """

    highlight_dluma = 0.05

    if len(palette) == 0:
        return
    if rows is None or columns is None:
        return

    cr.save()
    cr.translate(offset_x, offset_y)

    # Sizes and colors
    swatch_w = swatch_h = swatch_size
    light_col = HCYColor(color=bg_color)
    light_col.y += highlight_dluma
    dark_col = HCYColor(color=bg_color)
    dark_col.y -= highlight_dluma
    if light_col.y >= 1:
        light_col.y = 1.0
        dark_col.y = 1.0 - (2 * highlight_dluma)
    if dark_col.y <= 0:
        dark_col.y = 0.0
        light_col.y = 0.0 + (2 * highlight_dluma)

    # Upper left outline (bottom right is covered below by the
    # individual chips' shadows)
    ul_col = HCYColor(color=bg_color)
    ul_col.y *= 0.75
    ul_col.c *= 0.5
    cr.set_line_join(cairo.LINE_JOIN_ROUND)
    cr.set_line_cap(cairo.LINE_CAP_ROUND)
    cr.set_source_rgb(*ul_col.get_rgb())
    cr.move_to(0.5, (rows * swatch_h) - 1)
    cr.line_to(0.5, 0.5)
    row1cells = min(columns, len(palette))  # needed?
    cr.line_to((row1cells * swatch_w) - 1, 0.5)
    cr.set_line_width(2)
    cr.stroke()

    # Draw into the predefined grid
    r = c = 0
    cr.set_line_width(1.0)
    cr.set_line_cap(cairo.LINE_CAP_SQUARE)
    for col in palette.iter_colors():
        s_x = c * swatch_w
        s_y = r * swatch_h
        s_w = swatch_w
        s_h = swatch_h

        # Select fill bg and pattern fg colors, Tango-style edge highlight
        # and lower-right shadow.
        if col is None:
            # Empty slot, fill with a pattern
            hi_rgb = light_col.get_rgb()
            fill_bg_rgb = dark_col.get_rgb()
            fill_fg_rgb = light_col.get_rgb()
            sh_col = HCYColor(color=bg_color)
            sh_col.y *= 0.75
            sh_col.c *= 0.5
            sh_rgb = sh_col.get_rgb()
        else:
            # Color swatch
            hi_col = HCYColor(color=col)
            hi_col.y = min(hi_col.y * 1.1, 1)
            hi_col.c = min(hi_col.c * 1.1, 1)
            sh_col = HCYColor(color=col)
            sh_col.y *= 0.666
            sh_col.c *= 0.5
            hi_rgb = hi_col.get_rgb()
            fill_bg_rgb = col.get_rgb()
            fill_fg_rgb = None
            sh_rgb = sh_col.get_rgb()

        # Draw the swatch / color chip
        cr.set_source_rgb(*sh_rgb)
        cr.rectangle(s_x, s_y, s_w, s_h)
        cr.fill()
        cr.set_source_rgb(*fill_bg_rgb)
        cr.rectangle(s_x, s_y, s_w - 1, s_h - 1)
        cr.fill()
        if fill_fg_rgb is not None:
            s_w2 = int((s_w - 1) // 2)
            s_h2 = int((s_h - 1) // 2)
            cr.set_source_rgb(*fill_fg_rgb)
            cr.rectangle(s_x, s_y, s_w2, s_h2)
            cr.fill()
            cr.rectangle(s_x + s_w2, s_y + s_h2, s_w2, s_h2)
            cr.fill()
        cr.set_source_rgb(*hi_rgb)
        cr.rectangle(s_x + 0.5, s_y + 0.5, s_w - 2, s_h - 2)
        cr.stroke()

        c += 1
        if c >= columns:
            c = 0
            r += 1

    cr.restore()


## Module testing

if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    import sys
    win = Gtk.Window()
    win.set_title("palette view")
    win.connect("destroy", lambda *a: Gtk.main_quit())
    mgr = ColorManager(prefs={}, datapath=".")
    spv = PaletteView()
    spv.grid.show_matched_color = True
    spv.grid.can_select_empty = True
    spv.set_color_manager(mgr)
    spv.set_size_request(150, 150)
    if len(sys.argv[1:]) > 0:
        palette_file = sys.argv[1]  # GIMP palette file (*.gpl)
        palette = Palette(filename=palette_file)
        mgr.palette.update(palette)
    win.add(spv)
    win.show_all()
    Gtk.main()
